"""
This module contains classes and such that are responsible for generating
trajectories for MM/PBSA based on input, starting trajectories. It will
strip trajectories and output the ones that will be used for the calculations.
Needed for proper operation of gmx_MMPBSA

Methods:
         make_trajectories(INPUT, FILES, size): Makes all of the non-mutant
            trajectories needed for the calculation, as well as all dummy
            files (i.e. restarts and PDBs)

         make_mutant_trajectories(INPUT, FILES, rank): Mutates the trajectories

Classes:
         Trajectory: Class for manipulating Amber trajectories through cpptraj
"""

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

from warnings import warn
from GMXMMPBSA.exceptions import (TrajError, MMPBSA_Error, InternalError,
                                  MutantResError)


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def make_trajectories(INPUT, FILES, size, cpptraj, pre):
    """
    This function creates the necessary trajectory files, and creates thread-
    specific trajectories for parallel calculations
    """

    stability = FILES.stability

    # File suffix is dependent on file type
    if INPUT['netcdf']:
        trj_suffix = 'nc'
    else:
        trj_suffix = 'mdcrd'

    # Make sure we set up for a solvated topology; image, strip solvent, etc.
    # if FILES.solvated_prmtop:
    #    traj = Trajectory(FILES.solvated_prmtop, FILES.complex_trajs, cpptraj)
    #    traj.Setup(INPUT['startframe'], INPUT['endframe'], INPUT['interval'])
    #    if not stability:
    #       traj.Image()
    #    traj.StripSolvent(INPUT['strip_mask'])
    # # If we do not have a solvated topology...
    # else:
    #    print FILES.complex_prmtop, FILES.complex_trajs, cpptraj

    traj = Trajectory(FILES.complex_prmtop, FILES.complex_trajs, cpptraj)
    traj.Setup(INPUT['startframe'], INPUT['endframe'], INPUT['interval'])
    # RMS fit
    traj.rms('!(%s)' % INPUT['strip_mask'])

    num_frames_total = traj.processed_frames
    num_frames_nmode = 0

    # Sanity check
    if traj.processed_frames < size:
        raise MMPBSA_Error('Must have at least as many frames as processors!')

    # We now know how many frames we have in total, so make a list that lists the
    # number of frames found for each rank, and assign extra frames incrementally
    frames_per_rank = traj.processed_frames // size
    extras = traj.processed_frames - frames_per_rank * size
    frame_count = [frames_per_rank for i in range(size)]
    for i in range(size):
        if i < extras: frame_count[i] += 1

    # Dump our complex trajectories
    if INPUT['full_traj'] or INPUT['entropy'] == 1:
        traj.Outtraj(pre + 'complex.%s' % trj_suffix, filetype=INPUT['netcdf'])
    traj.Outtraj(pre + 'complex.pdb', frames='1', filetype='pdb')
    traj.Outtraj(pre + 'dummycomplex.inpcrd', frames='1', filetype='restart')

    # Now dump thread-specific trajectories
    last_frame = 1
    for i in range(size):
        frame_string = '%d-%d' % (last_frame, last_frame + frame_count[i] - 1)
        traj.Outtraj(pre + 'complex.%s.%d' % (trj_suffix, i),
                     frames=frame_string, filetype=INPUT['netcdf'])
        last_frame += frame_count[i]

    # Now create the receptor/ligand trajectories if we're taking them from
    # the complex trajectory

    if not stability and not FILES.receptor_trajs:
        traj.Strip(INPUT['ligand_mask'])
        if INPUT['full_traj'] or INPUT['entropy'] == 1:
            traj.Outtraj(pre + 'receptor.%s' % trj_suffix, filetype=INPUT['netcdf'])
        traj.Outtraj(pre + 'receptor.pdb', frames='1', filetype='pdb')
        traj.Outtraj(pre + 'dummyreceptor.inpcrd', frames='1', filetype='restart')
        last_frame = 1
        for i in range(size):
            frame_string = '%d-%d' % (last_frame, last_frame + frame_count[i] - 1)
            traj.Outtraj(pre + 'receptor.%s.%d' % (trj_suffix, i),
                         frames=frame_string, filetype=INPUT['netcdf'])
            last_frame += frame_count[i]
        traj.Unstrip(restrip_solvent=True)
        traj.rms('!(%s)' % INPUT['strip_mask'])

    if not stability and not FILES.ligand_trajs:
        traj.Strip(INPUT['receptor_mask'])
        if INPUT['full_traj'] or INPUT['entropy'] == 1:
            traj.Outtraj(pre + 'ligand.%s' % trj_suffix, filetype=INPUT['netcdf'])
        traj.Outtraj(pre + 'ligand.pdb', frames='1', filetype='pdb')
        traj.Outtraj(pre + 'dummyligand.inpcrd', frames='1', filetype='restart')
        last_frame = 1
        for i in range(size):
            frame_string = '%d-%d' % (last_frame, last_frame + frame_count[i] - 1)
            traj.Outtraj(pre + 'ligand.%s.%d' % (trj_suffix, i),
                         frames=frame_string, filetype=INPUT['netcdf'])
            last_frame += frame_count[i]
        traj.Unstrip(restrip_solvent=True)
        traj.rms('!(%s)' % INPUT['strip_mask'])

    # Run cpptraj to get the trajectory
    traj.Run(pre + 'normal_traj_cpptraj.out')

    # Go back and do the receptor and ligand if we used a multiple
    # trajectory approach
    if not stability and FILES.receptor_trajs:
        # Different actions depending on solvation...
        # if FILES.solvated_receptor_prmtop:
        #    rectraj = Trajectory(FILES.solvated_receptor_prmtop,
        #                         FILES.receptor_trajs, cpptraj)
        #    rectraj.Setup(INPUT['startframe'],INPUT['endframe'],INPUT['interval'])
        #    rectraj.StripSolvent(INPUT['strip_mask'])
        # else:
        rectraj = Trajectory(FILES.receptor_prmtop,
                             FILES.receptor_trajs, cpptraj)
        rectraj.Setup(INPUT['startframe'], INPUT['endframe'], INPUT['interval'])
        rectraj.rms('!(%s)' % INPUT['strip_mask'])
        if INPUT['full_traj'] or INPUT['entropy'] == 1:
            rectraj.Outtraj(pre + 'receptor.%s' % trj_suffix,
                            filetype=INPUT['netcdf'])
        rectraj.Outtraj(pre + 'receptor.pdb', frames='1', filetype='pdb')
        rectraj.Outtraj(pre + 'dummyreceptor.inpcrd', frames='1',
                        filetype='restart')

        # Now do the same split-up of workload as we did for complex, but don't
        # assume the same number of frames as we had for the complex
        if rectraj.processed_frames < size:
            raise MMPBSA_Error('Too many procs for receptor snapshots')
        frames_per_rank = rectraj.processed_frames // size
        extras = rectraj.processed_frames - frames_per_rank * size
        frame_count = [frames_per_rank for i in range(size)]
        for i in range(size):
            if i < extras: frame_count[i] += 1
        last_frame = 1
        for i in range(size):
            frame_string = '%d-%d' % (last_frame, last_frame + frame_count[i] - 1)
            rectraj.Outtraj(pre + 'receptor.%s.%d' % (trj_suffix, i),
                            frames=frame_string, filetype=INPUT['netcdf'])
            last_frame += frame_count[i]

        rectraj.Run(pre + 'receptor_traj_cpptraj.out')

    # end if not stability and FILES.receptor_trajs

    if not stability and FILES.ligand_trajs:
        # Different actions depending on solvation...
        # if FILES.solvated_ligand_prmtop:
        #    ligtraj = Trajectory(FILES.solvated_ligand_prmtop,
        #                         FILES.ligand_trajs, cpptraj)
        #    ligtraj.Setup(INPUT['startframe'],INPUT['endframe'],INPUT['interval'])
        #    ligtraj.StripSolvent(INPUT['strip_mask'])
        # else:
        ligtraj = Trajectory(FILES.ligand_prmtop, FILES.ligand_trajs, cpptraj)
        ligtraj.Setup(INPUT['startframe'], INPUT['endframe'], INPUT['interval'])
        ligtraj.rms('!(%s)' % INPUT['strip_mask'])
        ligtraj.Outtraj(pre + 'ligand.%s' % trj_suffix, filetype=INPUT['netcdf'])
        ligtraj.Outtraj(pre + 'ligand.pdb', frames='1', filetype='pdb')
        ligtraj.Outtraj(pre + 'dummyligand.inpcrd', frames='1',
                        filetype='restart')

        # Now do the same split-up of workload as we did for complex, but don't
        # assume the same number of frames as we had for the complex
        if ligtraj.processed_frames < size:
            raise MMPBSA_Error('Too many procs for ligand snapshots')
        frames_per_rank = ligtraj.processed_frames // size
        extras = ligtraj.processed_frames - frames_per_rank * size
        frame_count = [frames_per_rank for i in range(size)]
        for i in range(size):
            if i < extras: frame_count[i] += 1
        last_frame = 1
        for i in range(size):
            frame_string = '%d-%d' % (last_frame, last_frame + frame_count[i] - 1)
            ligtraj.Outtraj(pre + 'ligand.%s.%d' % (trj_suffix, i),
                            frames=frame_string, filetype=INPUT['netcdf'])
            last_frame += frame_count[i]

        ligtraj.Run(pre + 'ligand_traj_cpptraj.out')

    # end if not stability and FILES.ligand_trajs

    # Now make the nmode trajectories
    if INPUT['nmoderun']:
        nmtraj = Trajectory(FILES.complex_prmtop, [pre + 'complex.%s.%d' %
                                                   (trj_suffix, i) for i in range(size)], cpptraj)
        nmtraj.Setup(INPUT['nmstartframe'], INPUT['nmendframe'],
                     INPUT['nminterval'])

        num_frames_nmode = nmtraj.processed_frames

        # Now split up the complex trajectory by thread
        if nmtraj.processed_frames < size:
            raise MMPBSA_Error('More processors than complex nmode frames!')

        frames_per_rank = nmtraj.processed_frames // size
        extras = nmtraj.processed_frames - frames_per_rank * size
        frame_count = [frames_per_rank for i in range(size)]
        for i in range(size):
            if i < extras: frame_count[i] += 1
        last_frame = 1
        for i in range(size):
            frame_string = '%d-%d' % (last_frame, last_frame + frame_count[i] - 1)
            nmtraj.Outtraj(pre + 'complex_nm.%s.%d' % (trj_suffix, i),
                           frames=frame_string, filetype=INPUT['netcdf'])
            last_frame += frame_count[i]

        nmtraj.Run(pre + 'com_nm_traj_cpptraj.out')

        if not stability:
            nmtraj = Trajectory(FILES.receptor_prmtop, [pre + 'receptor.%s.%d' %
                                                        (trj_suffix, i) for i in range(size)], cpptraj)
            nmtraj.Setup(INPUT['nmstartframe'], INPUT['nmendframe'],
                         INPUT['nminterval'])

            # Now split up the complex trajectory by thread
            if nmtraj.processed_frames < size:
                raise MMPBSA_Error('More processors than receptor nmode frames!')

            frames_per_rank = nmtraj.processed_frames // size
            extras = nmtraj.processed_frames - frames_per_rank * size
            frame_count = [frames_per_rank for i in range(size)]
            for i in range(size):
                if i < extras: frame_count[i] += 1
            last_frame = 1
            for i in range(size):
                frame_string = '%d-%d' % (last_frame, last_frame + frame_count[i] - 1)
                nmtraj.Outtraj(pre + 'receptor_nm.%s.%d' % (trj_suffix, i),
                               frames=frame_string, filetype=INPUT['netcdf'])
                last_frame += frame_count[i]

            nmtraj.Run(pre + 'rec_nm_traj_cpptraj.out')

            nmtraj = Trajectory(FILES.ligand_prmtop, [pre + 'ligand.%s.%d' %
                                                      (trj_suffix, i) for i in range(size)], cpptraj)
            nmtraj.Setup(INPUT['nmstartframe'], INPUT['nmendframe'],
                         INPUT['nminterval'])

            # Now split up the complex trajectory by thread
            if nmtraj.processed_frames < size:
                raise MMPBSA_Error('More processors than ligand nmode frames!')

            frames_per_rank = nmtraj.processed_frames // size
            extras = nmtraj.processed_frames - frames_per_rank * size
            frame_count = [frames_per_rank for i in range(size)]
            for i in range(size):
                if i < extras: frame_count[i] += 1
            last_frame = 1
            for i in range(size):
                frame_string = '%d-%d' % (last_frame, last_frame + frame_count[i] - 1)
                nmtraj.Outtraj(pre + 'ligand_nm.%s.%d' % (trj_suffix, i),
                               frames=frame_string, filetype=INPUT['netcdf'])
                last_frame += frame_count[i]

            nmtraj.Run(pre + 'lig_nm_traj_cpptraj.out')

        # end if not stability

    # end if INPUT['nmoderun']

    return num_frames_total, num_frames_nmode


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

def make_mutant_trajectories(INPUT, FILES, rank, cpptraj,
                             norm_sys, mut_sys, pre):
    """ Mutates given trajectories and outputs dummy files for mutants """
    from GMXMMPBSA.alamdcrd import MutantMdcrd, GlyMutantMdcrd
    import shutil
    if not INPUT['alarun']: return None, None

    stability = FILES.stability

    if INPUT['netcdf']:
        trj_suffix = 'nc'
        raise TypeError('Alanine/Glycine scanning requires ASCII trajectories '
                        '(netcdf=0)')
    else:
        trj_suffix = 'mdcrd'

    if not stability and not ((FILES.ligand_prmtop == FILES.mutant_ligand_prmtop
                               and FILES.receptor_prmtop != FILES.mutant_receptor_prmtop) or (
                                      FILES.ligand_prmtop != FILES.mutant_ligand_prmtop and
                                      FILES.receptor_prmtop == FILES.mutant_receptor_prmtop)):
        raise MMPBSA_Error('Alanine/Glycine scanning requires either a mutated '
                           'ligand or receptor topology file with only 1 mutant residue, but not '
                           'both')

    master = rank == 0

    # Have each rank mutate our rank's normal complex trajectory
    try:
        com_mut = MutantMdcrd(pre + 'complex.%s.%d' % (trj_suffix, rank),
                              norm_sys.complex_prmtop, mut_sys.complex_prmtop)
    except MutantResError:
        com_mut = GlyMutantMdcrd(pre + 'complex.%s.%d' % (trj_suffix, rank),
                                 norm_sys.complex_prmtop, mut_sys.complex_prmtop)
    com_mut.MutateTraj(pre + 'mutant_complex.%s.%d' % (trj_suffix, rank))

    # Have each rank mutate our rank's normal receptor or ligand trajectory
    # and copy the normal one to the mutant if the mutated residue is *not*
    # present in there
    if not stability:
        if FILES.receptor_prmtop != FILES.mutant_receptor_prmtop:
            try:
                rec_mut = MutantMdcrd(pre + 'receptor.%s.%d' % (trj_suffix, rank),
                                      norm_sys.receptor_prmtop, mut_sys.receptor_prmtop)
            except MutantResError:
                rec_mut = GlyMutantMdcrd(pre + 'receptor.%s.%d' % (trj_suffix, rank),
                                         norm_sys.receptor_prmtop, mut_sys.receptor_prmtop)
            rec_mut.MutateTraj(pre + 'mutant_receptor.%s.%d' % (trj_suffix, rank))
            shutil.copyfile(pre + 'ligand.%s.%d' % (trj_suffix, rank),
                            pre + 'mutant_ligand.%s.%d' % (trj_suffix, rank))

        elif FILES.ligand_prmtop != FILES.mutant_ligand_prmtop:
            try:
                lig_mut = MutantMdcrd(pre + 'ligand.%s.%d' % (trj_suffix, rank),
                                      norm_sys.ligand_prmtop, mut_sys.ligand_prmtop)
            except MutantResError:
                lig_mut = GlyMutantMdcrd(pre + 'ligand.%s.%d' % (trj_suffix, rank),
                                         norm_sys.ligand_prmtop, mut_sys.ligand_prmtop)
            lig_mut.MutateTraj(pre + 'mutant_ligand.%s.%d' % (trj_suffix, rank))
            shutil.copyfile(pre + 'receptor.%s.%d' % (trj_suffix, rank),
                            pre + 'mutant_receptor.%s.%d' % (trj_suffix, rank))

    # Have our master dump out dummy files
    if master:
        com_traj = Trajectory(FILES.mutant_complex_prmtop,
                              pre + 'mutant_complex.%s.0' % trj_suffix, cpptraj)
        com_traj.Setup(1, 1, 1)
        com_traj.Outtraj(pre + 'mutant_complex.pdb', frames='1', filetype='pdb')
        com_traj.Outtraj(pre + 'mutant_dummycomplex.inpcrd', frames='1',
                         filetype='restart')
        com_traj.Run(pre + 'mutant_complex_cpptraj.out')
        if not stability:
            rec_traj = Trajectory(FILES.mutant_receptor_prmtop,
                                  pre + 'mutant_receptor.%s.0' % trj_suffix, cpptraj)
            rec_traj.Setup(1, 1, 1)
            rec_traj.Outtraj(pre + 'mutant_receptor.pdb', frames='1',
                             filetype='pdb')
            rec_traj.Outtraj(pre + 'mutant_dummyreceptor.inpcrd', frames='1',
                             filetype='restart')
            rec_traj.Run(pre + 'mutant_receptor_cpptraj.out')

            lig_traj = Trajectory(FILES.mutant_ligand_prmtop,
                                  pre + 'mutant_ligand.%s.0' % trj_suffix, cpptraj)
            lig_traj.Setup(1, 1, 1)
            lig_traj.Outtraj(pre + 'mutant_ligand.pdb', frames='1',
                             filetype='pdb')
            lig_traj.Outtraj(pre + 'mutant_dummyligand.inpcrd', frames='1',
                             filetype='restart')
            lig_traj.Run(pre + 'mutant_ligand_cpptraj.out')

    # Mutate our nmode trajectories if need be
    if INPUT['nmoderun']:
        com_mut = MutantMdcrd(pre + 'complex_nm.%s.%d' % (trj_suffix, rank),
                              norm_sys.complex_prmtop, mut_sys.complex_prmtop)
        com_mut.MutateTraj(pre + 'mutant_complex_nm.%s.%d' % (trj_suffix, rank))
        if not stability and FILES.receptor_prmtop != FILES.mutant_receptor_prmtop:
            rec_mut = MutantMdcrd(pre + 'receptor_nm.%s.%d' % (trj_suffix, rank),
                                  norm_sys.receptor_prmtop, mut_sys.receptor_prmtop)
            rec_mut.MutateTraj(pre + 'mutant_receptor_nm.%s.%d' %
                               (trj_suffix, rank))
            shutil.copyfile(pre + 'ligand_nm.%s.%d' % (trj_suffix, rank),
                            pre + 'mutant_ligand_nm.%s.%d' % (trj_suffix, rank))

        if not stability and FILES.ligand_prmtop != FILES.mutant_ligand_prmtop:
            lig_mut = MutantMdcrd(pre + 'ligand_nm.%s.%d' % (trj_suffix, rank),
                                  norm_sys.ligand_prmtop, mut_sys.ligand_prmtop)
            lig_mut.MutateTraj(pre + 'mutant_ligand_nm.%s.%d' %
                               (trj_suffix, rank))
            shutil.copyfile(pre + 'ligand_nm.%s.%d' % (trj_suffix, rank),
                            pre + 'mutant_ligand_nm.%s.%d' % (trj_suffix, rank))

    # If we're doing a quasi-harmonic approximation we need the full com traj
    if (INPUT['full_traj'] or INPUT['entropy'] == 1) and master:
        com_mut = MutantMdcrd(pre + 'complex.%s' % trj_suffix,
                              norm_sys.complex_prmtop, mut_sys.complex_prmtop)
        com_mut.MutateTraj(pre + 'mutant_complex.%s' % trj_suffix)

    return str(com_mut), com_mut.mutres


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Trajectory(object):
    """ Base Trajectory class:
        Methods: __init__(prmtop, traj_files)
                 Setup(startframe, endframe, interval)
                 Query()
                 Image()
                 Strip(mask)
                 StripSolvent(strip_mask)
                 Unstrip(restrip_solvent=True)
                 Outtraj(fname, startframe, endframe, interval, filetype, nobox)
                 Run(output)

        The way this class is typically used is initializing it by giving it a
        list of trajectory files. It will then automatically run Query to make
        sure that all trajectory files are valid, and it counts how many frames
        are present in each trajectory. Then you can strip the solvent (this
        just tags what is the solvent in case you want to re-strip it after every
        Unstrip() command). Then, add masks to strip out. Then, you can run with
        the current actions via the Run() routine. Typical sequence:

        trajsystem = Trajectory(solvated_prmtop, mdcrds)
        trajsystem.Setup(startframe, endframe, interval)
        trajsystem.Image()
        trajsystem.StripSolvent(strip_mask)
        trajsystem.Strip(receptor_mask)
        trajsystem.Outtraj('_MMPBSA_ligand.mdcrd', filetype=INPUT['netcdf'])
        trajsystem.Unstrip(restrip_solvent=True)
        trajsystem.Strip(ligand_mask)
        trajsystem.Outtraj('_MMPBSA_receptor.mdcrd', filetype=INPUT['netcdf'])
        trajsystem.Run('_MMPBSA_create_trajectories.out')
    """

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __init__(self, prmtop, traj_files, cpptraj='cpptraj'):
        """ Sets up a basic trajectory """

        # Make sure we have a list or string, and force it to be a list
        if isinstance(traj_files, list):
            self.traj_files = traj_files
        elif isinstance(traj_files, tuple):
            self.traj_files = list(traj_files)
        else:
            self.traj_files = [str(traj_files)]

        self.prmtop = prmtop

        # Find cpptraj
        self.exe = cpptraj

        self.strip_solvent = False

        self.Query()

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def Setup(self, startframe=1, endframe=9999999, interval=1):
        """ Finds the program and clears the queue and other attributes assigned
            during other function calls
        """

        self.strip_solvent = False
        orig_endframe = endframe
        orig_startframe = startframe

        # Delete the stripmask, but make its non-existence non-fatal, since it
        # won't exist unless Setup is called a second time.

        try:
            del self.stripmask
        except AttributeError:
            pass

        self.actions = []

        # Catch stupid choices
        if startframe > endframe:
            raise TrajError('Cannot have startframe (%d) > endframe (%d)' %
                            (startframe, endframe))

        if startframe < 0:
            raise TrajError('Startframe (%d) < 0' % startframe)

        if interval <= 0:
            raise TrajError('Interval (%d) <= 0' % interval)

        if startframe > self.total_frames:
            raise TrajError('start frame (%d) > total frames (%d)' %
                            (startframe, self.total_frames))

        # Bring endframe down if it's > the total number of frames

        endframe = min(endframe, self.total_frames)

        self.analyzed_frames = int((endframe - startframe) / interval) + 1

        # If we have an interval != 1, then it's possible that the endframe that
        # the user set is not the actual one that will be used. To make things
        # simpler, I will adjust the endframe to what is *actually* used

        endframe = startframe + interval * int((endframe - startframe) // interval)

        # Set up start and end arrays (for each trajectory)

        traj_starts = [-1 for i in range(len(self.traj_files))]
        traj_ends = [0 for i in range(len(self.traj_files))]

        # Now determine where each trajectory starts and ends

        for i in range(len(self.traj_files)):

            # skip over any trajectories that lie entirely before startframe

            if startframe > self.traj_sizes[i]:
                traj_starts[i] = -1  # this will tell us to skip this traj file
                startframe -= self.traj_sizes[i]
                continue

            # Now we start at our startframe

            traj_starts[i] = startframe

            # Now we figure out our last frame, adjusting for interval

            last_frame = startframe + interval * int((self.traj_sizes[i]
                                                      - startframe) / interval)

            traj_ends[i] = min(endframe, last_frame, self.traj_sizes[i])

            # Now determine our new start frame for the next trajectory. First
            # we need to know how many frames are left over at the end of the
            # last frame, and subtract those from the interval to determine
            # where our next trajectory should start from. Also move our endframe
            # down, but only by our last_frame! (since the last couple frames in
            # our trajectory file could be leftover due to interval > 1)

            startframe = interval + last_frame - self.traj_sizes[i]

            endframe -= last_frame

            if endframe < self.traj_sizes[i]:
                break  # we're done now

        # end for i len(traj_files)

        # We now have to trajin each of the analyzed trajectories

        for i in range(len(self.traj_files)):
            if traj_starts[i] < 0: continue  # skip -1's

            self.actions.append('trajin %s %d %d %d' % (self.traj_files[i],
                                                        traj_starts[i], traj_ends[i], interval))

        self.actions.append('noprogress')  # quash the progress bar

        self.processed_frames = (min(orig_endframe, self.total_frames) -
                                 orig_startframe) / interval + 1

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def StripSolvent(self, stripmask):
        """ Strips the solvent """
        self.stripmask = stripmask
        self.strip_solvent = True
        self.Strip(stripmask)

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def Query(self):
        """ Finds out how many frames are in the given traj files """
        from subprocess import Popen, PIPE
        import re

        framere = re.compile(r'Frames: (\d+)')

        self.traj_sizes = []

        # Determine how many frames are in each trajectory of the list

        for traj in self.traj_files:

            process = Popen([self.exe, '-p', str(self.prmtop), '-y', traj, '-tl'],
                            stdin=PIPE, stdout=PIPE)

            (output, error) = process.communicate(b'')

            if process.wait():  # if it quits with return code != 0
                raise TrajError('%s failed when querying %s' % (self.exe, traj))

            output = output.decode()

            # Now parse the output to find out how many frames are there. We are
            # looking for "Coordinate processing will occur on x frames."

            num_frames = framere.findall(output)
            if len(num_frames) < 1:
                raise TrajError('Could not find number of frames in ' + traj)
            elif len(num_frames) > 1:
                raise RuntimeError('Unexpected output from cpptraj. Has format '
                                   'changed?')
            self.traj_sizes.append(int(num_frames[0]))

        # end for traj in self.traj_files

        self.total_frames = sum(self.traj_sizes)

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def Image(self):
        """ Images the trajectories """
        from parmed.amber import LoadParm

        solvated_prmtop = LoadParm(self.prmtop)

        ifbox = solvated_prmtop.ptr('ifbox')
        if ifbox == 0:
            warn('Solvated topology %s has IFBOX == 0' % ifbox)

        self.actions.append('autoimage')

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def rms(self, mask):
        """ Does an RMS fit around a specific mask """
        self.actions.append('rmsd %s mass first' % mask)

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def Strip(self, mask):
        """ Strips a mask from the coordinates. """
        if mask is None or len(str(mask).strip()) == 0:
            raise InternalError('Cannot pass no mask to Strip!')
        self.actions.append('strip %s' % mask)

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def Unstrip(self, restrip_solvent=True):
        """ Returns to unstripped state """
        self.actions.append('unstrip')
        if self.strip_solvent and restrip_solvent:
            self.actions.append('strip %s' % self.stripmask)

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def Run(self, output_file=None):
        """ Runs cpptraj to actually create the files """
        from sys import stdout as sys_stdout
        from subprocess import Popen, PIPE

        # Accept output_file as both a file object and string object
        own_handle = False
        if output_file is None:
            stdout = sys_stdout
        else:
            try:
                stdout = open(output_file, 'w')
                own_handle = True
            except TypeError:
                stdout = output_file

        # Now it's time to run the program

        try:
            input_string = ''
            for action in self.actions:
                input_string += action.strip() + '\n'

            process = Popen([self.exe, self.prmtop], stdout=stdout, stdin=PIPE)

            process.communicate(input_string.encode())

            if process.wait():
                raise TrajError('Error running %s' % self.program)
        finally:
            if own_handle: stdout.close()

    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    def Outtraj(self, filename, frames=None, filetype='', nobox='nobox'):
        """ This adds an outtraj command to the action stack, and you can specify
            the type of trajectory file to output (such as restart/pdb for input
            files, etc.)
        """
        if not frames: frames = '1-%d' % self.total_frames
        self.actions.append('outtraj %s onlyframes %s %s %s' % (filename, frames,
                                                                nobox, filetype))

# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+



# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

import sys
from os.path import split
import shutil
from pathlib import Path
import logging
from PyQt5.QtWidgets import QApplication
# logFormatter = logging.Formatter("%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
# rootLogger = logging.getLogger(__name__)
# logging.getLogger(__name__).addHandler(logging.StreamHandler(sys.stdout))
logging.getLogger(__name__)
log_file = Path('gmx_MMPBSA.log')
if log_file.exists():
    log_file.unlink()

logging.basicConfig(
    level=logging.INFO,
    format="[%(levelname)-7s] %(message)s",
    handlers=[
        logging.FileHandler("gmx_MMPBSA.log"),
        logging.StreamHandler()])

try:
    from GMXMMPBSA.exceptions import GMXMMPBSA_ERROR, InputError, CommandlineError
    from GMXMMPBSA.infofile import InfoFile
    from GMXMMPBSA import main
    from GMXMMPBSA.analyzer import GMX_MMPBSA_ANA
    from GMXMMPBSA.commandlineparser import anaparser
except ImportError:
    import os
    amberhome = os.getenv('AMBERHOME') or '$AMBERHOME'
    logging.error('Could not import Amber Python modules. Please make sure '
                      'you have sourced %s/amber.sh (if you are using sh/ksh/'
                      'bash/zsh) or %s/amber.csh (if you are using csh/tcsh)' %
                      (amberhome, amberhome))
    raise ImportError('Could not import Amber Python modules. Please make sure '
                      'you have sourced %s/amber.sh (if you are using sh/ksh/'
                      'bash/zsh) or %s/amber.csh (if you are using csh/tcsh)' %
                      (amberhome, amberhome))


def gmxmmpbsa():

    # Adapted to run with MPI ?
    if len(sys.argv) > 1 and sys.argv[1] in ['MPI', 'mpi']:
        args = sys.argv
        args.pop(1)  # remove mpi arg before passed it to app
        try:
            from mpi4py import MPI
        except ImportError:
            GMXMMPBSA_ERROR('Could not import mpi4py package! Use serial version or install mpi4py.')
    else:
        # If we're not running "gmx_MMPBSA MPI", bring MPI into the top-level namespace
        # (which will overwrite the MPI from mpi4py, which we *want* to do in serial)
        from GMXMMPBSA.fake_mpi import MPI
        args = sys.argv
    # Set up error/signal handlers
    main.setup_run()

    # Instantiate the main MMPBSA_App
    app = main.MMPBSA_App(MPI)

    # Read the command-line arguments
    try:
        app.get_cl_args(args[1:])
    except CommandlineError as e:
        sys.stderr.write('%s: %s' % (type(e).__name__, e) + '\n')
        sys.exit(1)
    logging.info('Started')

    # Perform our MMPBSA --clean now
    if app.FILES.clean:
        sys.stdout.write('Cleaning temporary files and quitting.\n')
        app.remove(0)
        sys.exit(0)

    # See if we wanted to print out our input file options
    if app.FILES.infilehelp:
        app.input_file.print_contents(sys.stdout)
        sys.exit(0)

    # If we're not rewriting output do whole shebang, otherwise load info and parms
    # Throw up a barrier before and after running the actual calcs
    if not app.FILES.rewrite_output:
        try:
            app.read_input_file()
        except InputError as e:
            sys.stderr.write('%s: %s' % (type(e).__name__, e) + '\n')
            sys.stderr.write('  Enter `%s --help` for help\n' %
                             (split(sys.argv[0])[1]))
            sys.exit(1)
        app.process_input()
        app.check_for_bad_input()
        app.loadcheck_prmtops()
        app.file_setup()
        app.run_mmpbsa()
    # If we are rewriting output, load the info and check prmtops
    else:
        info = InfoFile(app)
        info.read_info()
        app.loadcheck_prmtops()

    # Now we parse the output, print, and finish
    app.parse_output_files()
    app.write_final_outputs()
    app.finalize()



def gmxmmpbsa_ana():
    app = QApplication(sys.argv)
    try:
        parser = anaparser.parse_args(sys.argv[1:])
    except CommandlineError as e:
        sys.stderr.write('%s: %s' % (type(e).__name__, e) + '\n')
        sys.exit(1)
    path = Path(parser.path).absolute()
    if not path.exists():
        print('Path not found')
        sys.exit(1)
    app.setApplicationName('gmx_MMPBSA Analyzer (gmx_MMPBSA_ana)')
    w = GMX_MMPBSA_ANA(path.as_posix())
    w.show()
    sys.exit(app.exec())

if __name__ == '__main__':


    logging.info('Finished')

    # gmxmmpbsa()
    gmxmmpbsa_ana()


"""
This module contains classes that back up the state of the MM/PBSA calculation
so that future post-processing of the output files can be done without
re-supplying all of the information again.
"""

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

import re
import warnings

class InfoFile(object):
    """
    This is the class responsible for backing up all of the calculation
    information and details so the output files can be re-written or the data
    can be readied for post-processing with external scripts.

    This object is passed a copy of the main app MMPBSA_App class
    """

    # Make a list about which INPUT variables are editable
    EDITABLE_INFO_VARS = ['debug_printlevel', 'verbose',
                          'csv_format', 'dec_verbose']

    def __init__(self, app):
        """ Instantiate me """
        self.app = app

    def write_info(self, name=None):
        """ Writes the info file to the info name """
        if name is None:
            name = self.app.pre + 'info'
        outfile = open(name, 'w')
        # The data we have to write: INPUT, FILES, and the following attributes:
        # numframes, numframes_nmode, mpi_size (also recognize size),
        # input_file_text, mut_str

        # Start with INPUT (and the editable vars). Allow this to recognize INFO
        # files from the last version of gmx_MMPBSA
        outfile.write('# You can alter the variables below to change what info '
                      'is printed out\n')
        for var in InfoFile.EDITABLE_INFO_VARS:
            outfile.write("INPUT['%s'] = %s\n" % (var,
                                                  self.write_var(self.app.INPUT[var])))
        outfile.write('#\n# MODIFY NOTHING BELOW HERE, OR GET WHAT YOU DESERVE\n')
        for var in list(self.app.INPUT.keys()):
            if var in InfoFile.EDITABLE_INFO_VARS: continue
            outfile.write("INPUT['%s'] = %s\n" % (var,
                                                  self.write_var(self.app.INPUT[var])))

        # Now print out the FILES
        for var in dir(self.app.FILES):
            # Skip over __method__ functions and output files
            if var.startswith('_') or var in ('rewrite_output', 'output_file',
                                              'decompout', 'energyout', 'dec_energies', 'overwrite'):
                continue
            outfile.write("FILES.%s = %s\n" % (var,
                                               self.write_var(getattr(self.app.FILES, var))))

        # Now print out the attributes we need to save
        outfile.write('size = %d\n' % self.app.mpi_size)
        outfile.write('numframes = %d\n' % self.app.numframes)
        outfile.write('numframes_nmode = %d\n' % self.app.numframes_nmode)
        outfile.write("mut_str = '%s'\n" % self.app.mut_str)
        outfile.write('using_chamber = %s\n' % self.app.using_chamber)
        outfile.write(self.app.input_file_text)

    def read_info(self, name=None):
        """ Reads a _MMPBSA_info file and populates the app """
        from GMXMMPBSA.commandlineparser import OptionList
        # Give self.app a FILES attribute if it does not have one yet before
        # trying to modify it
        if not hasattr(self.app, 'FILES'):
            self.app.FILES = OptionList()
        if name is None:
            name = self.app.pre + 'info'
        #TODO Add checking for differences in FILES
        inputre = re.compile(r'''INPUT\['(\S+)'\] = ((?:'.*')|(?:".*")|\S+)''')
        filesre = re.compile(r'''FILES\.(\S+) = (.*)''')
        otherre = re.compile(r'(\S+) = (\S+)')
        infile = open(name, 'r')
        input_text = ''
        for line in infile:
            if line.startswith('#'):
                continue
            # Now load each info file line
            if line.startswith('|'):
                input_text += line
                continue
            rematch = inputre.match(line)
            if rematch:
                var, val = rematch.groups()
                val = _determine_type(val)
                self.app.INPUT[var] = val
                continue
            rematch = filesre.match(line)
            if rematch:
                var, val = rematch.groups()
                val = _determine_type(val)
                setattr(self.app.FILES, var, val)
                continue
            rematch = otherre.match(line)
            if rematch:
                var, val = rematch.groups()
                val = _determine_type(val)
                if var == 'size':
                    self.app.mpi_size = val
                else:
                    setattr(self.app, var, val)
                continue
        # Determine stability here:
        self.app.stability = self.app.FILES.stability
        # Set app.pre as prefix
        self.app.pre = self.app.FILES.prefix
        if self.app.FILES.receptor_trajs or self.app.FILES.ligand_trajs:
            self.app.traj_protocol = 'MTP'
        else:
            self.app.traj_protocol = 'STP'
        # Load the input file text
        self.app.input_file_text = input_text

    def write_var(self, var):
        """
        Wrapper to return a string in which str vars are enclosed in quotes and
        numeric types (int and float) are not
        """
        if isinstance(var, str):
            return "'%s'" % var
        else:
            return "%s" % var


def _determine_type(thing):
    """ Determines what type this thing is """
    # If it is a string with quotes, strip off the quotes
    if thing.startswith('"') and thing.endswith('"'):
        return thing[1:len(thing)-1]
    if thing.startswith("'") and thing.endswith("'"):
        return thing[1:len(thing)-1]

    # Check for bool or None
    if thing == 'True':
        return True
    elif thing == 'False':
        return False
    elif thing == 'None':
        return None

    # Check for int, then check for float
    try:
        return int(thing)
    except ValueError:
        pass

    try:
        return float(thing)
    except ValueError:
        pass

    # Check for list
    if thing.startswith('[') and thing.endswith(']'):
        return eval(thing)

    # No idea what else it could be! Return string, but warn
    warnings.warn('Encountered unknown type in info file.\n' +
                  str(thing))
    return thing



"""
This module contains calculation classes that call the necessary programs
for running MM/PBSA calculations.

Methods:
   run_calculations(FILES, INPUT, rank) : Determines which calculations need to
        be run, then sets up the calculations and runs them

Classes:
   Calculation: Base calculation class
   EnergyCalculation: Typical GB/PB binding FE calculations. Handles all sander
                      and mmpbsa_py_energy program calls
   RISMCalculation: RISM binding FE calculation
   NmodeCalc: normal mode entropy calculation
   QuasiHarmCalc: Quasi-harmonic entropy calculation
"""

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

from GMXMMPBSA.exceptions import CalcError
from GMXMMPBSA.exceptions import GMXMMPBSA_ERROR
import os
import sys
import numpy as np
import math

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class CalculationList(list):
    """ This contains the list of all calculations that need to be run """

    def __init__(self, timer, *args):
        self.timer = timer
        self.timer_keys = []
        self.labels = []
        list.__init__(self)

    def append(self, calc, label='', timer_key=None):
        """ Add a new Calculation instance to the list """
        if not isinstance(calc, Calculation):
            raise TypeError('CalculationList can only take Calculation instances!')

        self.timer_keys.append(timer_key)
        list.append(self, calc)
        self.labels.append(label)

    def extend(self, calcs, labels, timer_keys):
        """ Add a list/iterable of Calculation instances to the list """
        for i, calc in enumerate(calcs):
            CalculationList.append(self, calc, labels[i], timer_keys[i])

    def run(self, rank, stdout=sys.stdout, stderr=sys.stderr):
        """ Runs every calculation in the list """
        own_handle = False
        try:
            f = open(stdout, 'w')
            own_handle = True
        except TypeError:
            f = stdout
        try:
            for i, calc in enumerate(self):
                # Start timer, run calculation, then stop the timer
                if self.timer_keys[i] is not None:
                    self.timer.start_timer(self.timer_keys[i])
                if self.labels[i]:
                    f.write(self.labels[i] + '\n')
                calc.setup()
                calc.run(rank, stdout=stdout, stderr=stderr)
                if self.timer_keys[i] is not None:
                    self.timer.stop_timer(self.timer_keys[i])
        finally:
            if own_handle: f.close()

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Calculation(object):
    """ Base calculation class. All other calculation classes should be inherited
        from this class.
    """

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __init__(self, prog, prmtop, incrd, inptraj, input_file, output):
        self.prmtop = str(prmtop)
        self.incrd = incrd
        self.input_file = input_file
        self.inptraj = inptraj
        self.output = output
        self.program = prog

        self.calc_setup = False # This means that the setup has run successfully

        self.command_args = [self.program]

    def run(self, rank, stdout=sys.stdout, stderr=sys.stderr):
        """ Runs the program. All command-line arguments must be set before
            calling this method. Command-line arguments should be set in setup()
        """
        from subprocess import Popen

        # If this has not been set up yet
        # then raise a stink
        if not self.calc_setup:
            raise CalcError('Cannot run a calculation without calling its' +
                            ' its setup() function!')

            # Here, make sure that we could pass a file *OR* a string as stdout/stderr.
        # If they are strings, then open files up with that name, and make sure to
        # close them afterwards. The setup() method should make sure that they are
        # either a file or a string!
        own_handleo = own_handlee = False
        try:
            process_stdout = open(stdout, 'w')
            own_handleo = True
        except TypeError:
            process_stdout = stdout
        try:
            process_stderr = open(stderr, 'w')
            own_handlee = True
        except TypeError:
            process_stderr = stderr

        # The setup() method sets the command-line arguments and makes sure that
        # all of the CL arguments are set. Now all we have to do is start the
        # process and monitor it for success.

        # Popen can only take strings as command-line arguments, so convert
        # everything to a string here. And if it appears to need the rank
        # substituted into the file name, substitute that in here
        try:
            for i in range(len(self.command_args)):
                self.command_args[i] = str(self.command_args[i])
                if '%d' in self.command_args[i]:
                    self.command_args[i] = self.command_args[i] % rank

            process = Popen(self.command_args, stdin=None, stdout=process_stdout,
                            stderr=process_stderr)

            calc_failed = bool(process.wait())

            if calc_failed:
                raise CalcError('%s failed with prmtop %s!' % (self.program,
                                                               self.prmtop))
        finally:
            if own_handleo: process_stdout.close()
            if own_handlee: process_stdout.close()

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def setup(self):
        """ Sets up the Calculation. Finds the program and adds that to the
            first element of the array. Inherited classes should call this
            method first, but then do anything else that is necessary for that
            calculation.
        """
        self.calc_setup= True

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class EnergyCalculation(Calculation):
    """ Uses mmpbsa_py_energy to evaluate energies """
    def __init__(self, prog, prmtop, incrd, inptraj, input_file, output, restrt):
        Calculation.__init__(self, prog, prmtop, incrd, inptraj,
                             input_file, output)
        self.restrt = restrt

    def setup(self):
        """
        Sets up the command-line arguments. Sander requires a unique restrt file
        for the MPI version (since one is *always* written and you don't want 2
        threads fighting to write the same dumb file)
        """
        self.command_args.append('-O')                    # overwrite flag
        self.command_args.extend(('-i', self.input_file)) # input file flag
        self.command_args.extend(('-p', self.prmtop))     # prmtop flag
        self.command_args.extend(('-c', self.incrd))      # input coordinate flag
        self.command_args.extend(('-y', self.inptraj))    # input trajectory flag
        self.command_args.extend(('-o', self.output))     # output file flag
        if self.restrt is not None:
            self.command_args.extend(('-r', self.restrt))  # restart file flag

        # Now test to make sure that the input file exists, since that's the only
        # one that may be absent (due to the use of -use-mdins)
        if not os.path.exists(self.input_file):
            raise IOError("Input file (%s) doesn't exist" % self.input_file)

        self.calc_setup = True

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class RISMCalculation(Calculation):
    """ This class handles RISM calculations """

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __init__(self, prog, prmtop, incrd, inptraj, xvvfile, output, INPUT):
        """ Sets up a RISM calculation. It's not as similar to the base class as
            other calculation classes are, but it still inherits useful methods
        """
        # rism3d.snglpnt dumps its output to stdout
        Calculation.__init__(self, prog, prmtop, incrd, inptraj, None, output)

        # Set up instance variables
        self.xvvfile      = xvvfile
        self.closure      = INPUT['closure']
        self.polardecomp  = INPUT['polardecomp']
        self.ng           = INPUT['ng'].replace(' ','') # get rid of spaces
        self.solvbox      = INPUT['solvbox']
        self.buffer       = INPUT['buffer']
        self.grdspc       = str(INPUT['grdspc']).replace(' ','')
        self.solvcut      = INPUT['solvcut']
        self.tolerance    = INPUT['tolerance']
        self.verbose      = INPUT['rism_verbose']
        self.solvbox      = INPUT['solvbox']
        self.gf           = INPUT['rismrun_gf']

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def setup(self):
        """ Sets up the RISM calculation. All it has to do is fill in the
            necessary command-line arguments
        """
        # Set up some defaults

        if self.ng == "-1,-1,-1":
            ngflag = False
        else:
            ngflag = True

        if self.solvbox == "-1,-1,-1":
            solvboxflag = False
        else:
            solvboxflag = True

        if self.polardecomp:
            polardecompflag = True
        else:
            polardecompflag = False

        Calculation.setup(self)
        self.command_args.extend( ('--xvv', self.xvvfile,
                                   '--closure', self.closure,
                                   '--buffer', self.buffer,
                                   '--grdspc', self.grdspc,
                                   '--solvcut', self.solvcut,
                                   '--tolerance', self.tolerance,
                                   '--verbose', self.verbose,
                                   '--prmtop', self.prmtop,
                                   '--pdb', self.incrd,
                                   '--traj', self.inptraj))
        if ngflag:
            self.command_args.extend(('--ng', self.ng))
        if solvboxflag:
            self.command_args.extend(('--solvbox', self.solvbox))
        if polardecompflag:
            self.command_args.extend(['--polarDecomp'])
        if self.gf:
            self.command_args.extend(['--gf'])
        if not os.path.exists(self.xvvfile):
            raise IOError('XVVFILE (%s) does not exist!' % self.xvvfile)

        self.calc_setup = True

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def run(self, rank, *args, **kwargs):
        Calculation.run(self, rank, stdout=self.output % rank)

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class NmodeCalc(Calculation):
    """ Calculates entropy contribution by normal mode approximation """

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __init__(self, prog, prmtop, incrd, inptraj, output, INPUT):
        """ Initializes the nmode calculation. Need to set the options string """
        from math import sqrt
        Calculation.__init__(self, prog, prmtop, incrd, inptraj, None, output)

        kappa = sqrt(0.10806 * INPUT['nmode_istrng'])
        if INPUT['nmode_igb']:
            option_string = ('ntpr=10000, diel=C, kappa=%f, cut=1000, gb=1, ' +
                             'dielc=%f, temp0=%f') % (kappa,
                                                      INPUT['dielc'], INPUT['temp'])
        else:
            option_string = ('ntpr=10000, diel=R, kappa=%f, cut=1000, gb=0, ' +
                             'dielc=%f, temp0=%f') % (kappa,
                                                      INPUT['dielc'], INPUT['temp'])

        self.option_string = option_string
        self.drms = INPUT['drms']
        self.maxcyc = INPUT['maxcyc']

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def setup(self):
        """ Sets up the simulation """

        self.command_args.extend((self.incrd, self.prmtop, self.maxcyc, self.drms,
                                  self.option_string, self.inptraj))
        self.calc_setup = True

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def run(self, rank, *args, **kwargs):
        Calculation.run(self, rank, stdout=self.output % rank)

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class QuasiHarmCalc(Calculation):
    """ Quasi-harmonic entropy calculation class """

    def __init__(self, prog, prmtop, inptraj, input_file, output,
                 receptor_mask, ligand_mask, fnpre):
        """ Initializes the Quasi-harmonic calculation class """
        Calculation.__init__(self, prog, prmtop, None, inptraj,
                             input_file, output)
        self.stability = not bool(receptor_mask) and not bool(ligand_mask)
        self.receptor_mask, self.ligand_mask = receptor_mask, ligand_mask
        self.calc_setup = False
        self.fnpre = fnpre # file name prefix

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def setup(self):
        """ Sets up a Quasi-harmonic calculation """
        from subprocess import Popen, PIPE

        # Determine the prefix from our input file... hack way to do this
        if self.input_file.startswith(self.fnpre + 'mutant_'):
            prefix = self.fnpre + 'mutant_'
        else:
            prefix = self.fnpre

        # Make sure masks are a list, and that there are enough masks

        # First thing we need is the average PDB as a reference
        ptraj_str = 'trajin %s\naverage %savgcomplex.pdb pdb chainid " "\ngo' % (self.inptraj,
                                                                     prefix)

        outfile = open(self.fnpre + 'create_average.out','w')

        process = Popen([self.program, self.prmtop], stdin=PIPE, stdout=outfile)
        out, err = process.communicate(ptraj_str.encode())

        if process.wait():
            raise CalcError('Failed creating average PDB')

        outfile.close()

        # Now that we have the PDB file

        self.command_args.extend((self.prmtop, self.input_file))

        self.calc_setup = True

    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def run(self, rank, *args, **kwargs):
        Calculation.run(self, rank, stdout=self.output)

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class PBEnergyCalculation(EnergyCalculation):
    """
    Specially handle the PB calculations to extract warnings and errors PBSA
    prints to stdout and redirect them to the user
    """
    def run(self, rank, stdout=sys.stdout, stderr=sys.stderr):
        """
        Runs the program. All command-line arguments must be set before calling
        this method. Command-line arguments should be set in setup()
        stdout is ignored here because we need to parse it for errors
        """
        import re
        from subprocess import Popen, PIPE

        # If this has not been set up yet
        # then raise a stink
        if not self.calc_setup:
            raise CalcError('Cannot run a calculation without calling its' +
                            ' its setup() function!')

        errorre = re.compile('(pb (?:bomb)|(?:warning))', re.I)
        # Here, make sure that we could pass a file *OR* a string as stderr.
        own_handle = False
        try:
            process_stderr = open(stderr, 'w')
            own_handle = True
        except TypeError:
            process_stderr = stderr

        # The setup() method sets the command-line arguments and makes sure that
        # all of the CL arguments are set. Now all we have to do is start the
        # process and monitor it for success.

        # Popen can only take strings as command-line arguments, so convert
        # everything to a string here. If rank needs to be substituted in, do that
        # here
        try:
            for i in range(len(self.command_args)):
                self.command_args[i] = str(self.command_args[i])
                if '%d' in self.command_args[i]:
                    self.command_args[i] = self.command_args[i] % rank

            process = Popen(self.command_args, stdin=None, stdout=PIPE,
                            stderr=process_stderr)

            out, err = process.communicate(b'')
            calc_failed = bool(process.wait())
            out = out.decode('utf-8')
            if calc_failed:
                error_list = [s.strip() for s in out.split('\n')
                              if errorre.match(s.strip())]

                GMXMMPBSA_ERROR('%s failed with prmtop %s!\n\t' % (self.program, self.prmtop) +
                                '\n\t'.join(error_list) + '\n' +
                                'If you are using sander and PB calculation, check the *.mdout files to get the sander '
                                'error and check this thread (http://archive.ambermd.org/201303/0548.html)\n',
                                CalcError)
        finally:
            if own_handle: process_stderr.close()

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class SurfCalc(Calculation):
    """
    Base class for a surface area calculation using cpptraj. Children must
    implement _get_instring(self, rank) which returns the string containing the
    necessary cpptraj input
    """

    def __init__(self, prog, prmtop, inptraj, output, probe=1.4, offset=0.0):
        self.prmtop = str(prmtop)
        self.inptraj = inptraj
        self.output = output
        self.program = prog
        self.probe = probe
        self.offset = offset

    def run(self, rank, stdout=sys.stdout, stderr=sys.stderr):
        """ Runs the program. All command-line arguments must be set before
            calling this method. Command-line arguments should be set in setup()
        """
        from subprocess import Popen, PIPE

        # If this has not been set up yet
        # then raise a stink
        if not self.calc_setup:
            raise CalcError('Cannot run a calculation without calling its' +
                            ' its setup() function!')

            # Make sure the inptraj and output are rank-substituted
        instring = self._get_instring(rank)

        process = Popen([self.program, self.prmtop], stdin=PIPE, stdout=PIPE,
                        stderr=PIPE)

        out, err = process.communicate(instring.encode())

        calc_failed = bool(process.wait())

        if calc_failed:
            raise CalcError('%s failed with prmtop %s!' % (self.program,
                                                           self.prmtop))

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class LcpoCalc(SurfCalc):
    """
    Uses LCPO to calculate the surface area
    (Linear Combination of Pairwise Overlaps)
    """

    def _get_instring(self, rank):
        """ Returns the cpptraj input string """
        inptraj = self.inptraj % rank
        output = self.output % rank
        return "trajin %s\nsolvent none\nsurf :* out %s\n" % (inptraj, output)

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class MolsurfCalc(SurfCalc):
    """ Uses molsurf to calculate the surface area """

    def __init__(self, prog, prmtop, inptraj, output, probe=1.4, offset=0.0):
        SurfCalc.__init__(self, prog, prmtop, inptraj, output)
        self.probe = probe
        self.offset = offset

    def _get_instring(self, rank):
        inptraj = self.inptraj % rank
        output = self.output % rank
        return "trajin %s\nmolsurf :* out %s probe %s offset %s" % (inptraj,
                                                                    output, self.probe, self.offset)

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class CopyCalc(Calculation):
    """
    This is for mutant files that are unchanged from 'normal' files (i.e., when
    the mutation is in the receptor, the ligand outputs are copied)
    """
    def __init__(self, orig_name, final_name):
        self.orig_name = orig_name
        self.final_name = final_name

    def run(self, rank, stdout=None, stderr=None):
        from shutil import copy
        # Do rank-substitution if necessary
        if '%d' in self.orig_name:
            orig_name = self.orig_name % rank
        else:
            orig_name = self.orig_name

        if '%d' in self.final_name:
            final_name = self.final_name % rank
        else:
            final_name = self.final_name

        copy(orig_name, final_name)

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class PrintCalc(Calculation):
    """
    This is just a way to insert a printed message to the screen during the
    calculation list execution
    """
    def __init__(self, message):
        self.message = message

    def run(self, rank, stdout=sys.stdout, stderr=sys.stderr):
        if rank == 0: stdout.write(self.message + '\n')

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class InteractionEntropyCalc:
    """
    Class for Interaction Entropy calculation
    :return {IE_key: data}
    """
    def __init__(self, ggas, key, INPUT, output):
        self.ggas = ggas
        self.key = key
        self.INPUT = INPUT
        self.output = output
        self.data = {}

        self._calculate()
        self.save_output()

    def _calculate(self):
        # gases constant in kcal/mol
        k = 0.001987
        energy_int = np.array([], dtype=np.float)
        a_energy_int = np.array([], dtype=np.float)
        d_energy_int = np.array([], dtype=np.float)
        exp_energy_int = np.array([], dtype=np.float)
        ts = np.array([], dtype=np.float)

        for eint in self.ggas:
            energy_int = np.append(energy_int, eint)
            aeint = energy_int.mean()
            a_energy_int = np.append(a_energy_int, aeint)
            deint = eint - aeint
            d_energy_int = np.append(d_energy_int, deint)
            eceint = math.exp(deint / (k * self.INPUT['entropy_temp']))
            exp_energy_int = np.append(exp_energy_int, eceint)
            aeceint = exp_energy_int.mean()
            cts = k * self.INPUT['entropy_temp'] * math.log(aeceint)
            ts = np.append(ts, cts)
        self.IntEnt = ts
        self.data['IE_' + self.key] = self.IntEnt
        nframes = len(self.IntEnt)
        self.ie_frames = math.ceil(nframes * (self.INPUT['entropy_seg'] / 100))
        self.value = self.IntEnt[self.ie_frames:].mean()
        self.frames = [x for x in range(self.INPUT['startframe'], self.INPUT['endframe'] + self.INPUT['interval'],
                                       self.INPUT['interval'])]

    def save_output(self):
        with open(self.output, 'w') as out:
            out.write(f'Calculation for last {self.ie_frames} frames:\n')
            out.write(f'Interaction Entropy: {self.value}\n\n')
            out.write(f'Interaction Entropy per-frame:\n')

            out.write('Frame # | IE value\n')
            for f, d in zip(self.frames, self.IntEnt):
                out.write('{:d}  {:.2f}\n'.format(f,d))

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+



"""
This is a module that contains functions generally useful for the
gmx_MMPBSA script. A full list of functions/subroutines is shown below.
It must be included to insure proper functioning of gmx_MMPBSA

List of functions and a brief description of their purpose
-remove: Removes temporary work files in this directory. It has a number of
    different levels to remove only a small number of files.
-concatenate: combines 2 files into a single, common file
"""

# TODO get rid of this file altogether and move these functions into the main
# app class

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

import os
import shutil
from pathlib import Path
import json
import logging


def checkff(overwrite):
    """

    :param overwrite:
    :param sel_ff: folder/leaprc
    Examples:
        gmxMMPBSA/leaprc.GLYCAM_06h-1
        oldff/leaprc.ff99SB
        leaprc.protein.ff14SB

    :return:
    """
    amberhome = os.getenv('AMBERHOME')
    if not amberhome:
        logging.error('Could not found Amber. Please make sure you have sourced %s/amber.sh (if you are using sh/ksh/'
              'bash/zsh) or %s/amber.csh (if you are using csh/tcsh)' %
              (amberhome, amberhome))
        return
    amberhome = Path(amberhome)

    logging.info('Checking if supported force fields exists in Amber data...')
    if overwrite:
        logging.info('The overwrite_data option is enabled. Overwriting the gmxMMPBSA data...')
    data_path = Path(__file__).parent.joinpath('data')
    info_file = data_path.joinpath('info.dat')
    with info_file.open('r') as read_file:
        data = json.load(read_file)

    leap_dat = amberhome.joinpath('dat/leap/')

    for ff in data['forcefield']:
        for p in data['forcefield'][ff]:
            gmxf = leap_dat.joinpath(p, 'gmxMMPBSA')
            if not gmxf.exists():
                gmxf.mkdir()
            if type(data['forcefield'][ff][p]) == list:
                for l in data['forcefield'][ff][p]:
                    cf = data_path.joinpath(l)
                    if overwrite:
                        shutil.copy(cf, gmxf)
                    else:
                        if cf.exists() and not gmxf.joinpath(l).exists():
                            shutil.copy(cf, gmxf)
            else:
                cf = data_path.joinpath(data['forcefield'][ff][p])
                if overwrite:
                    shutil.copy(cf, gmxf)
                else:
                    if not cf.exists():
                        shutil.copy(cf, gmxf)
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def remove(flag, mpi_size=0, fnpre='_GMXMMPBSA_'):
    """ Removes temporary files. Allows for different levels of cleanliness """

    # A list of all input files that we keep for the flag -use-mdins
    input_files = [fnpre + 'gb.mdin', fnpre + 'pb.mdin', fnpre + 'pb.mdin2',
                   fnpre + 'gb_decomp_com.mdin', fnpre + 'gb_decomp_rec.mdin',
                   fnpre + 'gb_decomp_lig.mdin', fnpre + 'pb_decomp_com.mdin',
                   fnpre + 'pb_decomp_rec.mdin', fnpre + 'pb_decomp_lig.mdin',
                   fnpre + 'cpptrajentropy.in',
                   fnpre + 'mutant_cpptrajentropy.in',
                   fnpre + 'gb_qmmm_com.mdin', fnpre + 'gb_qmmm_rec.mdin',
                   fnpre + 'gb_qmmm_lig.mdin']

    # All the extra files we keep for keep_files = 1
    keep_files_1 = [fnpre + 'ligand.mdcrd', fnpre + 'ligand.nc',
                    fnpre + 'mutant_ligand.mdcrd', fnpre + 'mutant_ligand.nc',
                    fnpre + 'complex.mdcrd', fnpre + 'mutant_complex.mdcrd',
                    fnpre + 'complex.nc', fnpre + 'mutant_complex.nc',
                    fnpre + 'receptor.mdcrd', fnpre + 'mutant_receptor.mdcrd',
                    fnpre + 'receptor.nc', fnpre + 'mutant_receptor.nc',
                    fnpre + 'dummycomplex.inpcrd', fnpre + 'complex.pdb',
                    fnpre + 'dummyreceptor.inpcrd', fnpre + 'receptor.pdb',
                    fnpre + 'dummyligand.inpcrd', fnpre + 'ligand.pdb',
                    fnpre + 'mutant_dummycomplex.inpcrd',
                    fnpre + 'mutant_dummyreceptor.inpcrd',
                    fnpre + 'mutant_dummyligand.inpcrd',
                    fnpre + 'mutant_complex.pdb', fnpre + 'mutant_receptor.pdb',
                    fnpre + 'mutant_ligand.pdb', fnpre + 'complex_nm.mdcrd',
                    fnpre + 'complex_nm.nc', fnpre + 'mutant_complex_nm.mdcrd',
                    fnpre + 'mutant_complex_nm.nc', fnpre + 'receptor_nm.mdcrd',
                    fnpre + 'receptor_nm.nc', fnpre + 'mutant_receptor_nm.nc',
                    fnpre + 'mutant_receptor_nm.mdcrd', fnpre + 'ligand_nm.nc',
                    fnpre + 'ligand_nm.mdcrd', fnpre + 'mutant_ligand_nm.nc',
                    fnpre + 'mutant_ligand_nm.mdcrd', fnpre + 'avgcomplex.pdb',
                    fnpre + 'mutant_avgcomplex.pdb', fnpre + 'ligand_entropy.out',
                    fnpre + 'complex_entropy.out', fnpre + 'receptor_entropy.out',
                    fnpre + 'cpptraj_entropy.out',
                    fnpre + 'mutant_cpptraj_entropy.out',
                    fnpre + 'mutant_complex_entropy.out',
                    fnpre + 'mutant_receptor_entropy.out',
                    fnpre + 'mutant_ligand_entropy.out',
                    fnpre + 'complex_gb.mdout', fnpre + 'mutant_complex_gb.mdout',
                    fnpre + 'receptor_gb.mdout', fnpre + 'mutant_receptor_gb.mdout',
                    fnpre + 'ligand_gb.mdout', fnpre + 'mutant_ligand_gb.mdout',
                    fnpre + 'complex_pb.mdout', fnpre + 'mutant_complex_pb.mdout',
                    fnpre + 'receptor_pb.mdout', fnpre + 'mutant_receptor_pb.mdout',
                    fnpre + 'ligand_pb.mdout', fnpre + 'mutant_ligand_pb.mdout',
                    fnpre + 'complex_rism.mdout',
                    fnpre + 'mutant_complex_rism.mdout',
                    fnpre + 'receptor_rism.mdout',
                    fnpre + 'mutant_receptor_rism.mdout',
                    fnpre + 'ligand_rism.mdout', fnpre + 'mutant_ligand_rism.mdout',
                    fnpre + 'complex_nm.out', fnpre + 'mutant_complex_nm.out',
                    fnpre + 'receptor_nm.out', fnpre + 'mutant_receptor_nm.out',
                    fnpre + 'ligand_nm.out', fnpre + 'mutant_ligand_nm.out',
                    fnpre + 'complex_gb_surf.dat', fnpre + 'receptor_gb_surf.dat',
                    fnpre + 'ligand_gb_surf.dat',
                    fnpre + 'mutant_complex_gb_surf.dat',
                    fnpre + 'mutant_receptor_gb_surf.dat',
                    fnpre + 'mutant_ligand_gb_surf.dat', fnpre + 'info']

    # Collect all of the temporary files (those starting with _MMPBSA_)
    allfiles = os.listdir(os.getcwd())
    tempfiles = []
    for fil in allfiles:
        if fil.startswith(fnpre): tempfiles.append(fil)

    if flag == -1:  # internal -- keep all mdin files
        for fil in tempfiles:
            if not fil in input_files: os.remove(fil)
    elif flag == 0:  # remove all temporary files
        for fil in tempfiles: os.remove(fil)
    elif flag == 1:  # keep keep mdcrds, mdouts, and other relevant output files
        for fil in tempfiles:
            if fil in keep_files_1: continue  # keep this file
            # Now we have to split out this file and analyze the base. If the
            # suffix is just a number (corresponding to a thread-specific output
            # file or trajectory), then we only want to remove it if in the base
            # name is not in keep_files_1
            base, ext = os.path.splitext(fil)
            if ext.strip('.').isdigit() and base in keep_files_1: continue
            # if we've made it this far, remove the file
            os.remove(fil)


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def concatenate(file1, file2):
    """ Adds contents of file2 onto beginning of file1 """
    import os
    chunksize = 1048576  # Read the file in 1 MB chunks
    # Open the 2 files, the first in append mode
    with open(file2, 'r') as fl2:
        # Add a newline (make it OS-independent) to the first file if it doesn't
        # already end in one
        file1.write(os.linesep)

        str1 = fl2.read(chunksize)
        while str1:
            file1.write(str1)
            str1 = fl2.read(chunksize)

    file1.flush()
    # Now remove the merged file (file2)
    os.remove(file2)


class Unbuffered(object):
    """ Takes a stream handle and calls flush() on it after writing """

    def __init__(self, handle):
        self._handle = handle

    def write(self, data):
        self._handle.write(data)
        self._handle.flush()

    def __getattr__(self, attr):
        return getattr(self._handle, attr)


class PDB:
    def __init__(self):

        self.allAtoms = []
        self.info = []
        self.element = None

    def parse(self, filename):
        ofile = open(filename).readlines()
        model = 0
        for line in ofile:
            if 'MODEL' in line:
                model += 1
                if model > 1:
                    raise ValueError('Only one Model is allowed')
            elif 'ATOM' in line[0:6] or 'HETATM' in line[0:6]:
                atm = {}
                atm['id'] = line[0:6].strip()
                atm["number"] = int(line[6:11].strip())
                atm["name"] = line[12:16].strip()
                atm["alternate_location"] = line[16].strip()
                atm["resname"] = line[17:21].strip()
                atm["chain"] = line[21].strip()
                atm["resnum"] = int(line[22:26].strip())
                atm["i_code"] = line[26].strip()
                atm["x"] = float(line[30:38].strip())
                atm["y"] = float(line[38:46].strip())
                atm["z"] = float(line[46:54].strip())
                atm["occupancy"] = float(line[54:60].strip())
                atm["b_factor"] = float(line[60:66].strip())
                atm['segment_id'] = line[72:76].strip()
                atm["element"] = line[76:78]
                atm["charge"] = line[78:80]
                self.allAtoms.append(atm)
            elif 'TER' in line[0:6]:
                atm = {}
                atm['id'] = 'TER'
                self.allAtoms.append(atm)
            elif 'ENDML' in line[0:6]:
                continue
            else:
                self.info.append(line)

    def getOutLine(self, atm, ter=False):
        # Create the PDB line.
        if ter:
            atomrec = '%-6s\n' % (atm['id'])
        else:
            atomrec = '%-6s%5d %-4s%1s%-3s %1s%4d%1s   %8.3f%8.3f%8.3f%6.2f%6.2f      %-4s%2s%-2s\n' % (
                atm['id'], atm["number"], atm["name"], atm["alternate_location"], atm["resname"], atm["chain"],
                atm["resnum"], atm["i_code"], atm["x"], atm["y"], atm["z"], atm["occupancy"], atm["b_factor"],
                atm['segment_id'], atm["element"], atm["charge"]
            )
        return atomrec



"""
This is a module that provides the same attributes as mpi4py so that
we can reduce the number of #ifdef MPIs that are used. Ideally, the
only ones that will be needed after using this module are the ones to
make sure that this is loaded #ifndef MPI and related ones.
"""

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

class MPI(object):
    """ Fake MPI class :) """

    class Communicator(object):

        #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

        def Get_rank(self):
            """ The rank of a serial process is always 0 """
            return 0

        #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

        def Get_size(self):
            """ The size of a serial version is always 1 """
            return 1

        #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

        def bcast(self, inp, root=0):
            """ Mimics an MPI_Bcast """
            return inp

        #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

        def Barrier(self):
            """ Mimics an MPI_Barrier """
            pass

        #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

        def Abort(self, status=1):
            """ Just exits """
            from sys import exit
            exit(status)

    @staticmethod
    def Finalize():
        return

    COMM_WORLD = Communicator()

#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+




# This file was generated by 'versioneer.py' (0.19) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2021-03-01T01:32:59-0500",
 "dirty": false,
 "error": null,
 "full-revisionid": "8f20334f0f5208f4f4b38a6998aadf821ae8bff6",
 "version": "v1.3.2"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



"""
This module provides a means for users to take advantage of gmx_MMPBSA's parsing
ability. It exposes the free energy data (optionally to numpy arrays) so that
users can write a simple script to carry out custom data analyses, leveraging
the full power of Python's extensions, if they want (e.g., numpy, scipy, etc.)
"""

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

from copy import deepcopy
from GMXMMPBSA import infofile, main, amber_outputs
from GMXMMPBSA.exceptions import SetupError, NoFileExists
from GMXMMPBSA.fake_mpi import MPI
import os
import warnings
# Try importing numpy to see if we have it available
try:
    import numpy as np
    HAS_NUMPY = True
except ImportError:
    from array import array
    HAS_NUMPY = False

__all__ = ['load_mmpbsa_info']

# Abstract the array-making routines
if HAS_NUMPY:
    make_array = lambda x: np.fromiter(x, float)
    make_array_len = lambda x: np.zeros(x, float)
else:
    make_array = lambda x: array('d', x)
    make_array_len = lambda x: array('d', [0 for i in range(x)])

class mmpbsa_data(dict):
    """ Main class that holds all of the Free Energy data """
    def __init__(self, app):
        """ Load data from an info object """
        global HAS_NUMPY
        if not isinstance(app, main.MMPBSA_App):
            raise TypeError('mmpbsa_data can only take an MMPBSA_App!')
        # Loop through all of the data
        if not hasattr(app, 'calc_types'):
            raise SetupError('Output files have not yet been parsed!')
        # Now load the data into the dict
        has_mutant = False
        # See if we are doing stability
        self.stability = app.stability
        # Now load the data
        for key in app.calc_types:
            if key == 'mutant' or key =='qh':
                has_mutant = True
                continue
            self[key] = {}
            tmpdict = {}
            for dkey in app.calc_types[key]['complex'].data:
                tmpdict[dkey] = make_array(app.calc_types[key]['complex'].data[dkey])
            self[key]['complex'] = tmpdict
            if not self.stability:
                tmpdict = {}
                for dkey in app.calc_types[key]['receptor'].data:
                    tmpdict[dkey] = make_array(app.calc_types[key]['receptor'].data[dkey])
                self[key]['receptor'] = tmpdict
                tmpdict = {}
                for dkey in app.calc_types[key]['ligand'].data:
                    tmpdict[dkey] = make_array(app.calc_types[key]['ligand'].data[dkey])
                self[key]['ligand'] = tmpdict
                tmpdict = {}
                for dkey in app.calc_types[key]['delta'].data:
                    tmpdict[dkey] = make_array(app.calc_types[key]['delta'].data[dkey])
                self[key]['delta'] = tmpdict

        # Are we doing a mutant?
        if has_mutant:
            self.mutant = {}
            for key in app.calc_types['mutant']:
                if key == 'qh': continue
                self.mutant[key] = {}
                tmpdict = {}
                for dkey in app.calc_types['mutant'][key]['complex'].data:
                    tmpdict[dkey] = make_array(
                        app.calc_types['mutant'][key]['complex'].data[dkey])
                self.mutant[key]['complex'] = tmpdict
                if not self.stability:
                    self.mutant[key]['receptor'] = {}
                    tmpdict = {}
                    for dkey in app.calc_types['mutant'][key]['receptor'].data:
                        tmpdict[dkey] = make_array(
                            app.calc_types['mutant'][key]['receptor'].data[dkey])
                    self.mutant[key]['receptor'] = tmpdict
                    tmpdict = {}
                    for dkey in app.calc_types['mutant'][key]['ligand'].data:
                        tmpdict[dkey] = make_array(
                            app.calc_types['mutant'][key]['ligand'].data[dkey])
                    self.mutant[key]['ligand'] = tmpdict
                    tmpdict = {}
                    for dkey in app.calc_types['mutant'][key]['delta'].data:
                        tmpdict[dkey] = make_array(
                            app.calc_types['mutant'][key]['delta'].data[dkey])
                    self.mutant[key]['delta'] = tmpdict


    def __iadd__(self, other):
        """
        Adding one to another extends every array. The way we do this depends on
        whether we're using numpy arrays or
        """
        global HAS_NUMPY
        if HAS_NUMPY:
            return self._add_numpy(other)
        else:
            return self._add_nonumpy(other)

    def _add_numpy(self, other):
        """
        If we have numpy available, we need to extend every array in a numpy-valid
        way
        """
        used_keys = []
        for key in self:
            used_keys.append(key)
            try:
                for dkey in self[key]['complex']:
                    _combine_np_arrays(self[key]['complex'][dkey],
                                       other[key]['complex'][dkey])
                for dkey in self[key]['receptor']:
                    _combine_np_arrays(self[key]['receptor'][dkey],
                                       other[key]['receptor'][dkey])
                for dkey in self[key]['ligand']:
                    _combine_np_arrays(self[key]['ligand'][dkey],
                                       other[key]['ligand'][dkey])
            except KeyError:
                pass
        for key in other:
            if key in used_keys:
                continue
            # If we didn't have a particular calc type, copy that array in here
            self[key] = deepcopy(other[key])
        # Check mutant statuses. If the other has mutant and I don't, copy other
        # If we both have mutant, combine.  If only I do, already done
        if self.mutant and not other.mutant:
            self.mutant = deepcopy(other.mutant)
        elif self.mutant and other.mutant:
            used_keys_mutant = []
            for key in self.mutant:
                used_keys_mutant.append(key)
                try:
                    for dkey in self[key]['complex']:
                        _combine_np_arrays(self[key]['complex'][dkey],
                                           other[key]['complex'][dkey])
                    for dkey in self[key]['receptor']:
                        _combine_np_arrays(self[key]['receptor'][dkey],
                                           other[key]['receptor'][dkey])
                    for dkey in self[key]['ligand']:
                        _combine_np_arrays(self[key]['ligand'][dkey],
                                           other[key]['ligand'][dkey])
                except KeyError:
                    pass
            for key in other.mutant:
                if key in used_keys_mutant:
                    continue
                self.mutant[key] = deepcopy(other.mutant[key])

    def _add_nonumpy(self, other):
        """ Adds up 2 array objects (just use 'extend' method) """
        used_keys = []
        for key in self:
            used_keys.append(key)
            for dkey in self[key]:
                try:
                    self[key]['complex'][dkey].extend(other[key]['complex'][dkey])
                    self[key]['receptor'][dkey].extend(other[key]['receptor'][dkey])
                    self[key]['ligand'][dkey].extend(other[key]['ligand'][dkey])
                except KeyError:
                    pass
        for key in other:
            if key in used_keys:
                continue
            # If we didn't have a particular calc type, copy that array in here
            self[key] = deepcopy(other[key])
        # Check mutant statuses. If the other has mutant and I don't, copy other
        # If we both have mutant, combine.  If only I do, already done
        if self.mutant and not other.mutant:
            self.mutant = deepcopy(other.mutant)
        elif self.mutant and other.mutant:
            used_keys_mutant = []
            for key in self.mutant:
                used_keys_mutant.append(key)
                for dkey in self.mutant[key]:
                    try:
                        self.mutant[key]['complex'].extend(
                            other.mutant[key]['complex'])
                        self.mutant[key]['receptor'].extend(
                            other.mutant[key]['receptor'])
                        self.mutant[key]['ligand'].extend(
                            other.mutant[key]['ligand'])
                    except KeyError:
                        pass
            for key in other.mutant:
                if key in used_keys_mutant:
                    continue
                self.mutant[key] = deepcopy(other.mutant[key])


def _combine_np_arrays(nparray1, nparray2):
    origsize = nparray1.shape[0]
    nparray1.resize(origsize + nparray2.shape[0])
    for i in range(nparray2.shape[0]):
        nparray1[origsize + i] = nparray2[i]

    def load_topologies(self):
        """
        Loads the topology files so we have residue information for decomp
        analyses
        """
        self.app.loadcheck_prmtops()

class APIDecompOut(amber_outputs.DecompOut):

    def __init__(self, basename, surften, num_files, verbose, nframes):
        amber_outputs.DecompOut.__init__(self, basename, None, surften, False,
                                         num_files, verbose)
        self.array_data = {}
        # Make a new dict for all printed tokens (TDC,SDC,BDC)
        for key in self.allowed_tokens:
            self.array_data[key] = {}
            for i in range(nframes):
                for j in range(self.num_terms):
                    rnum, internal, vdw, eel, pol, sas, tot = self.get_next_term(key)
                    try:
                        self.array_data[key][rnum]['int'][i] = internal
                        self.array_data[key][rnum]['vdw'][i] = vdw
                        self.array_data[key][rnum]['eel'][i] = eel
                        self.array_data[key][rnum]['pol'][i] = pol
                        self.array_data[key][rnum]['sas'][i] = sas
                        self.array_data[key][rnum]['tot'][i] = tot
                    except KeyError:
                        # This is the first frame, we don't have the rnum dict yet, so
                        # make that dict here and create the arrays, then fill the
                        # first term
                        self.array_data[key][rnum] = {}
                        for k in ('int', 'vdw', 'eel', 'pol', 'sas', 'tot'):
                            self.array_data[key][rnum][k] = make_array_len(nframes)
                        self.array_data[key][rnum]['int'][i] = internal
                        self.array_data[key][rnum]['vdw'][i] = vdw
                        self.array_data[key][rnum]['eel'][i] = eel
                        self.array_data[key][rnum]['pol'][i] = pol
                        self.array_data[key][rnum]['sas'][i] = sas
                        self.array_data[key][rnum]['tot'][i] = tot

class APIPairDecompOut(amber_outputs.PairDecompOut):

    def __init__(self, basename, surften, num_files, verbose, nframes):
        amber_outputs.DecompOut.__init__(self, basename, None, surften, False,
                                         num_files, verbose)
        self.array_data = {}
        # Make a new dict for all printed tokens (TDC,SDC,BDC)
        for key in self.allowed_tokens:
            self.array_data[key] = {}

        for i in range(nframes):
            for key in self.allowed_tokens:
                for j in range(self.num_terms):
                    rnum, rnum2, internal, vdw, eel, pol, sas, tot = \
                        self.get_next_term(key)
                    dkey = '%d-%d' % (rnum, rnum2)
                    try:
                        self.array_data[key][dkey]['int'][i] = internal
                        self.array_data[key][dkey]['vdw'][i] = vdw
                        self.array_data[key][dkey]['eel'][i] = eel
                        self.array_data[key][dkey]['pol'][i] = pol
                        self.array_data[key][dkey]['sas'][i] = sas
                        self.array_data[key][dkey]['tot'][i] = tot
                    except KeyError:
                        # This is the first frame, we don't have the rnum dict yet, so
                        # make that dict here and create the arrays, then fill the
                        # first term
                        self.array_data[key][dkey] = {}
                        for k in ('int', 'vdw', 'eel', 'pol', 'sas', 'tot'):
                            self.array_data[key][dkey][k] = make_array_len(nframes)
                        self.array_data[key][dkey]['int'][i] = internal
                        self.array_data[key][dkey]['vdw'][i] = vdw
                        self.array_data[key][dkey]['eel'][i] = eel
                        self.array_data[key][dkey]['pol'][i] = pol
                        self.array_data[key][dkey]['sas'][i] = sas
                        self.array_data[key][dkey]['tot'][i] = tot

class APIDecompOutGMX(amber_outputs.DecompOut):

    def __init__(self, basename, surften, num_files, verbose, nframes, lig_num=None):
        amber_outputs.DecompOut.__init__(self, basename, None, surften, False,
                                         num_files, verbose)
        self.array_data = {}
        # Make a new dict for all printed tokens (TDC,SDC,BDC)
        for key in self.allowed_tokens:
            self.array_data[key] = {}
        for i in range(nframes):
            for key in self.allowed_tokens:
                n = 0
                for j in range(self.num_terms):
                    rnum, internal, vdw, eel, pol, sas, tot = self.get_next_term(key)
                    if lig_num:
                        if n == len(lig_num):
                            n = 0
                        rnum = lig_num[n]
                        n += 1
                    if rnum not in self.array_data[key]:
                        self.array_data[key][rnum] = {}
                        for k in ('int', 'vdw', 'eel', 'pol', 'sas', 'tot'):
                            self.array_data[key][rnum][k] = make_array_len(nframes)
                    self.array_data[key][rnum]['int'][i] = internal
                    self.array_data[key][rnum]['vdw'][i] = vdw
                    self.array_data[key][rnum]['eel'][i] = eel
                    self.array_data[key][rnum]['pol'][i] = pol
                    self.array_data[key][rnum]['sas'][i] = sas
                    self.array_data[key][rnum]['tot'][i] = tot

class APIPairDecompOutGMX(amber_outputs.PairDecompOut):

    def __init__(self, basename, surften, num_files, verbose, nframes, lig_num=None):
        amber_outputs.DecompOut.__init__(self, basename, None, surften, False,
                                         num_files, verbose)
        self.array_data = {}
        # Make a new dict for all printed tokens (TDC,SDC,BDC)
        for key in self.allowed_tokens:
            self.array_data[key] = {}

        for i in range(nframes):
            for key in self.allowed_tokens:
                m = 0
                n = 0
                for j in range(self.num_terms):
                    rnum, rnum2, internal, vdw, eel, pol, sas, tot = self.get_next_term(key)
                    if lig_num: # rename lig residues like complex lig names
                        if n == len(lig_num):
                            m += 1
                            n = 0
                        rnum = lig_num[m]
                        rnum2 = lig_num[n]
                        n += 1
                    if rnum not in self.array_data[key]:
                        self.array_data[key][rnum] = {}
                    if rnum2 not in self.array_data[key][rnum]:
                        self.array_data[key][rnum][rnum2] =  {}
                        for k in ('int', 'vdw', 'eel', 'pol', 'sas', 'tot'):
                            self.array_data[key][rnum][rnum2][k] = make_array_len(nframes)
                    self.array_data[key][rnum][rnum2]['int'][i] = internal
                    self.array_data[key][rnum][rnum2]['vdw'][i] = vdw
                    self.array_data[key][rnum][rnum2]['eel'][i] = eel
                    self.array_data[key][rnum][rnum2]['pol'][i] = pol
                    self.array_data[key][rnum][rnum2]['sas'][i] = sas
                    self.array_data[key][rnum][rnum2]['tot'][i] = tot

def get_delta_decomp(app, decomp_calc_type, data):
    """

    :param calc_type: gb or pb
    :param data:
    :return:
    """
    data_out = {decomp_calc_type: {'delta': {}}}
    # data_out[decomp_calc_type] = {'delta': {}}
    tempdict = {}
    # complex, if stability: receptor, ligand
    com = data[decomp_calc_type]['complex']
    rec = data[decomp_calc_type]['receptor']
    lig = data[decomp_calc_type]['ligand']
    for p in com:
        tempdict[p] = {}
        for res in com[p]:
            if res not in tempdict[p]:
                tempdict[p][res] = {}
            if app.INPUT['idecomp'] in [1, 2]:
                if rec and res in rec[p]:
                    for para in com[p][res]:
                        d = com[p][res][para] - rec[p][res][para]
                        tempdict[p][res][para] = d
                elif lig and res in lig[p]:
                    for para in com[p][res]:
                        d = com[p][res][para] - lig[p][res][para]
                        tempdict[p][res][para] = d
            else:
                if rec and res in rec[p]:
                    for resp, resp1 in zip(com[p][res], rec[p][res]):
                        tempdict[p][res][resp] = {}
                        for para in com[p][res][resp]:
                            d = com[p][res][resp][para] - rec[p][res][resp1][para]
                            tempdict[p][res][resp][para] = d
                    for resp in com[p][res]:
                        tempdict[p][res][resp] = com[p][res][resp]
                elif lig and res in lig[p]:
                    for resp, resp1 in zip(list(com[p][res])[(len(com[p][res]) - len(lig[p][res])):],
                                           lig[p][res]):
                        tempdict[p][res][resp] = {}
                        for para in com[p][res][resp]:
                            d = com[p][res][resp][para] - lig[p][res][resp1][para]
                            tempdict[p][res][resp][para] = d
                    for resp in com[p][res]:
                        tempdict[p][res][resp] = com[p][res][resp]
    return tempdict

def load_mmpbsa_info(fname):
    """
    Loads up an MMPBSA.py info file and returns a mmpbsa_data instance with all
    of the data available in numpy arrays if numpy is available. The returned
    object is a mmpbsa_data instance.

    mmpbsa_data attributes:
    -----------------------
       o  Derived from "dict"
       o  Each solvent model is a dictionary key for a numpy array (if numpy is
          available) or array.array (if numpy is unavailable) for each of the
          species (complex, receptor, ligand) present in the calculation.
       o  The alanine scanning mutant data is under another dict denoted by the
          'mutant' key.

    Data Layout:
    ------------
       Solvent Model     |  Dictionary Key    |  Data Keys Available
       -------------------------------------------------------------------
       Generalized Born  |  'gb'              |  EGB, ESURF, *
       Poisson-Boltzmann |  'pb'              |  EPB, EDISPER, ECAVITY, *
       3D-RISM (GF)      |  'rism gf'         |
       3D-RISM (Standard)|  'rism std'        |
       Normal Mode       |  'nmode'           |
       Quasi-harmonic    |  'qh'              |

    * == TOTAL, VDW, EEL, 1-4 EEL, 1-4 VDW, BOND, ANGLE, DIHED

    The keys above are entries for the main dict as well as the sub-dict whose
    key is 'mutant' in the main dict.  Each entry in the main (and mutant sub-)
    dict is, itself, a dict with 1 or 3 keys; 'complex', 'receptor', 'ligand';
    where 'receptor' and 'ligand' are missing for stability calculations.
    If numpy is available, all data will be numpy.ndarray instances.  Otherwise,
    all data will be array.array instances.

    All of the objects referenced by the listed 'Dictionary Key's are dicts in
    which the listed 'Data Keys Available' are keys to the data arrays themselves

    Examples:
    ---------
       # Load numpy for our analyses (optional)
       import numpy as np

       # Load the _MMPBSA_info file:
       mydata = load_mmpbsa_info('_MMPBSA_info')

       # Access the complex GB data structure and calculate the autocorr. fcn.
       autocorr = np.correlate(mydata['gb']['complex']['TOTAL'],
                               mydata['gb']['complex']['TOTAL'])

       # Calculate the standard deviation of the alanine mutant receptor in PB
       print mydata.mutant['pb']['receptor']['TOTAL'].std()
    """
    if not HAS_NUMPY:
        warnings.warn('numpy was not found. Data will be packed in normal Python '
                      'arrays. Install numpy for more efficient array handling.',
                      NotImplemented)

    if not os.path.exists(fname):
        raise NoFileExists("cannot find %s!" % fname)
    app = main.MMPBSA_App(MPI)
    info = infofile.InfoFile(app)
    info.read_info(fname)
    app.normal_system = app.mutant_system = None
    app.parse_output_files()
    return_data = mmpbsa_data(app)
    # Since Decomp data is parsed in a memory-efficient manner (by not storing
    # all of the data in arrays, but rather by printing each data point as it's
    # parsed), we need to handle the decomp data separately here
    if app.INPUT['decomprun']:
        # Simplify the decomp class instance creation
        if app.INPUT['idecomp'] in (1, 2):
            DecompClass = lambda x, y: APIDecompOut(x, y, app.mpi_size,
                                                    app.INPUT['dec_verbose'], app.numframes)
        else:
            DecompClass = lambda x, y: APIPairDecompOut(x, y, app.mpi_size,
                                                        app.INPUT['dec_verbose'], app.numframes)

        if not app.INPUT['mutant_only']:
            # Do normal GB
            if app.INPUT['gbrun']:
                return_data['decomp'] = {'gb' : {}}
                return_data['decomp']['gb']['complex'] = DecompClass(
                    app.FILES.prefix + 'complex_gb.mdout',
                    app.INPUT['surften']).array_data
                if not app.stability:
                    return_data['decomp']['gb']['receptor'] = DecompClass(
                        app.FILES.prefix + 'receptor_gb.mdout',
                        app.INPUT['surften']).array_data
                    return_data['decomp']['gb']['ligand'] = DecompClass(
                        app.FILES.prefix + 'ligand_gb.mdout',
                        app.INPUT['surften']).array_data
            # Do normal PB
            if app.INPUT['pbrun']:
                return_data['decomp'] = {'pb' : {}}
                return_data['decomp']['pb']['complex'] = DecompClass(
                    app.FILES.prefix + 'complex_pb.mdout',
                    app.INPUT['surften']).array_data
                if not app.stability:
                    return_data['decomp']['pb']['receptor'] = DecompClass(
                        app.FILES.prefix + 'receptor_pb.mdout',
                        app.INPUT['surften']).array_data
                    return_data['decomp']['pb']['ligand'] = DecompClass(
                        app.FILES.prefix + 'ligand_pb.mdout',
                        app.INPUT['surften']).array_data
        if app.INPUT['alarun']:
            # Do mutant GB
            if app.INPUT['gbrun']:
                return_data.mutant['decomp'] = {'gb' : {}}
                return_data.mutant['decomp']['gb']['complex'] = DecompClass(
                    app.FILES.prefix + 'mutant_complex_gb.mdout',
                    app.INPUT['surften']).array_data
                if not app.stability:
                    return_data.mutant['decomp']['gb']['receptor'] = DecompClass(
                        app.FILES.prefix + 'mutant_receptor_gb.mdout',
                        app.INPUT['surften']).array_data
                    return_data.mutant['decomp']['gb']['ligand'] = DecompClass(
                        app.FILES.prefix + 'mutant_ligand_gb.mdout',
                        app.INPUT['surften']).array_data
            # Do mutant PB
            if app.INPUT['pbrun']:
                return_data.mutant['decomp'] = {'pb' : {}}
                return_data.mutant['decomp']['pb']['complex'] = DecompClass(
                    app.FILES.prefix + 'mutant_complex_pb.mdout',
                    app.INPUT['surften']).array_data
                if not app.stability:
                    return_data.mutant['decomp']['pb']['receptor'] = DecompClass(
                        app.FILES.prefix + 'mutant_receptor_pb.mdout',
                        app.INPUT['surften']).array_data
                    return_data.mutant['decomp']['pb']['ligand'] = DecompClass(
                        app.FILES.prefix + 'mutant_ligand_pb.mdout',
                        app.INPUT['surften']).array_data
        else:
            return_data.mutant = None

    return return_data

def load_gmxmmpbsa_info(fname):
    """
    Loads up an gmx_MMPBSA info file and returns a mmpbsa_data instance with all
    of the data available in numpy arrays if numpy is available. The returned
    object is a mmpbsa_data instance.

    change the structure to get more easy way to graph per residue

    mmpbsa_data attributes:
    -----------------------
       o  Derived from "dict"
       o  Each solvent model is a dictionary key for a numpy array (if numpy is
          available) or array.array (if numpy is unavailable) for each of the
          species (complex, receptor, ligand) present in the calculation.
       o  The alanine scanning mutant data is under another dict denoted by the
          'mutant' key.

    Data Layout:
    ------------
       Solvent Model     |  Dictionary Key    |  Data Keys Available
       -------------------------------------------------------------------
       Generalized Born  |  'gb'              |  EGB, ESURF, *
       Poisson-Boltzmann |  'pb'              |  EPB, EDISPER, ECAVITY, *
       3D-RISM (GF)      |  'rism gf'         |
       3D-RISM (Standard)|  'rism std'        |
       Normal Mode       |  'nmode'           |
       Quasi-harmonic    |  'qh'              |

    * == TOTAL, VDW, EEL, 1-4 EEL, 1-4 VDW, BOND, ANGLE, DIHED

    The keys above are entries for the main dict as well as the sub-dict whose
    key is 'mutant' in the main dict.  Each entry in the main (and mutant sub-)
    dict is, itself, a dict with 1 or 3 keys; 'complex', 'receptor', 'ligand';
    where 'receptor' and 'ligand' are missing for stability calculations.
    If numpy is available, all data will be numpy.ndarray instances.  Otherwise,
    all data will be array.array instances.

    All of the objects referenced by the listed 'Dictionary Key's are dicts in
    which the listed 'Data Keys Available' are keys to the data arrays themselves

    Examples:
    ---------
       # Load numpy for our analyses (optional)
       import numpy as np

       # Load the _MMPBSA_info file:
       mydata = load_mmpbsa_info('_MMPBSA_info')

       # Access the complex GB data structure and calculate the autocorr. fcn.
       autocorr = np.correlate(mydata['gb']['complex']['TOTAL'],
                               mydata['gb']['complex']['TOTAL'])

       # Calculate the standard deviation of the alanine mutant receptor in PB
       print mydata.mutant['pb']['receptor']['TOTAL'].std()
    """
    if not HAS_NUMPY:
        warnings.warn('numpy was not found. Data will be packed in normal Python '
                      'arrays. Install numpy for more efficient array handling.',
                      NotImplemented)

    if not os.path.exists(fname):
        raise NoFileExists("cannot find %s!" % fname)
    app = main.MMPBSA_App(MPI)
    info = infofile.InfoFile(app)
    info.read_info(fname)
    app.normal_system = app.mutant_system = None
    app.parse_output_files()
    return_data = mmpbsa_data(app)
    # Since Decomp data is parsed in a memory-efficient manner (by not storing
    # all of the data in arrays, but rather by printing each data point as it's
    # parsed), we need to handle the decomp data separately here
    if not app.INPUT['alarun']:
        return_data.mutant = None
    lig_res = None
    if app.INPUT['decomprun']:
        # Simplify the decomp class instance creation
        if app.INPUT['idecomp'] in (1, 2):
            DecompClass = lambda x, y, z=None: APIDecompOutGMX(x, y, app.mpi_size,
                                                    app.INPUT['dec_verbose'], app.numframes, lig_num=z)
        else:
            DecompClass = lambda x, y, z=None: APIPairDecompOutGMX(x, y, app.mpi_size,
                                                        app.INPUT['dec_verbose'], app.numframes, lig_num=z)

        if not app.INPUT['mutant_only']:
            # Do normal GB
            if app.INPUT['gbrun']:
                return_data['decomp'] = {'gb' : {}}
                return_data['decomp']['gb']['complex'] = DecompClass(app.FILES.prefix + 'complex_gb.mdout',
                                                                     app.INPUT['surften']).array_data
                com_res = []
                for k in return_data['decomp']['gb']['complex']['TDC']:
                    if k not in com_res:
                        com_res.append(k)

                if not app.stability:
                    return_data['decomp']['gb']['receptor'] = DecompClass(app.FILES.prefix + 'receptor_gb.mdout',
                                                                          app.INPUT['surften']).array_data
                    rec_res = []
                    for k in return_data['decomp']['gb']['receptor']['TDC']:
                        if k not in rec_res:
                            rec_res.append(k)
                    lig_res = com_res[len(rec_res):]
                    return_data['decomp']['gb']['ligand'] = DecompClass(app.FILES.prefix + 'ligand_gb.mdout',
                                                                        app.INPUT['surften'], lig_res).array_data
                    return_data['decomp']['gb']['delta'] = get_delta_decomp(app, 'gb', return_data['decomp'])
            # Do normal PB
            if app.INPUT['pbrun']:
                return_data['decomp'] = {'pb' : {}}
                return_data['decomp']['pb']['complex'] = DecompClass(app.FILES.prefix + 'complex_pb.mdout',
                                                                     app.INPUT['surften']).array_data
                if not app.stability:
                    return_data['decomp']['pb']['receptor'] = DecompClass(app.FILES.prefix + 'receptor_pb.mdout',
                                                                          app.INPUT['surften']).array_data
                    return_data['decomp']['pb']['ligand'] = DecompClass(app.FILES.prefix + 'ligand_pb.mdout',
                                                                        app.INPUT['surften'], lig_res).array_data
                    return_data['decomp']['pb']['delta'] = get_delta_decomp(app, 'pb', return_data['decomp'])
        if app.INPUT['alarun']:
            # Do mutant GB
            if app.INPUT['gbrun']:
                return_data.mutant['decomp'] = {'gb' : {}}
                return_data.mutant['decomp']['gb']['complex'] = DecompClass(app.FILES.prefix +
                                                                             'mutant_complex_gb.mdout',
                                                                            app.INPUT['surften']).array_data
                if not app.stability:
                    return_data.mutant['decomp']['gb']['receptor'] = DecompClass(app.FILES.prefix +
                                                                                  'mutant_receptor_gb.mdout',
                                                                                 app.INPUT['surften']).array_data
                    return_data.mutant['decomp']['gb']['ligand'] = DecompClass(app.FILES.prefix +
                                                                                'mutant_ligand_gb.mdout',
                                                                               app.INPUT['surften'], lig_res).array_data
                    return_data.mutant['decomp']['gb']['delta'] = get_delta_decomp(app, 'gb',
                                                                                   return_data.mutant['decomp'])
            # Do mutant PB
            if app.INPUT['pbrun']:
                return_data.mutant['decomp'] = {'pb' : {}}
                return_data.mutant['decomp']['pb']['complex'] = DecompClass(app.FILES.prefix +
                                                                             'mutant_complex_pb.mdout',
                                                                            app.INPUT['surften']).array_data
                if not app.stability:
                    return_data.mutant['decomp']['pb']['receptor'] = DecompClass(app.FILES.prefix +
                                                                                  'mutant_receptor_pb.mdout',
                                                                                 app.INPUT['surften']).array_data
                    return_data.mutant['decomp']['pb']['ligand'] = DecompClass(app.FILES.prefix +
                                                                                'mutant_ligand_pb.mdout',
                                                                               app.INPUT['surften'], lig_res).array_data
                    return_data.mutant['decomp']['pb']['delta'] = get_delta_decomp(app, 'pb',
                                                                                   return_data.mutant['decomp'])
        else:
            return_data.mutant = None

    return return_data, app




"""
This is a module that contains functions responsible for parsing the
input file for gmx_MMPBSA. It must be included with gmx_MMPBSA to
ensure proper functioning.
"""

# ##############################################################################
#                           GPLv3 LICENSE INFO                                 #
#                                                                              #
#  Copyright (C) 2020  Mario S. Valdes-Tresanco and Mario E. Valdes-Tresanco   #
#  Copyright (C) 2014  Jason Swails, Bill Miller III, and Dwight McGee         #
#                                                                              #
#   Project: https://github.com/Valdes-Tresanco-MS/gmx_MMPBSA                  #
#                                                                              #
#   This program is free software; you can redistribute it and/or modify it    #
#  under the terms of the GNU General Public License version 3 as published    #
#  by the Free Software Foundation.                                            #
#                                                                              #
#  This program is distributed in the hope that it will be useful, but         #
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    #
#  for more details.                                                           #
# ##############################################################################

from GMXMMPBSA.exceptions import InputError, InternalError


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Variable(object):
    """ Base variable class. It has a name and a single value """

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __init__(self, varname, dat_type=int, default=None, chars_to_match=4,
                 case_sensitive=False, description=''):
        """ Initializes the variable type. Sets the default value as well as
          specifying how many characters are required by the parser to trigger
          recognition
      """
        # Catch illegalities
        if not dat_type in (int, str, float):
            raise InputError('Variable has unknown data type %s' %
                             dat_type.__name__)

        # You can't match more characters than you have characters!
        chars_to_match = min(chars_to_match, len(varname))

        self.name = varname
        self.datatype = dat_type
        if default is not None:
            if self.datatype is str:
                self.value = default.replace("'", '').replace('"', '')
            else:
                self.value = self.datatype(default)
        else:
            self.value = None
        self.tomatch = chars_to_match
        self.case_sensitive = case_sensitive
        self.description = description

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __str__(self):
        """ Prints statistics for the variable """
        string = 'Variable name:  %s\n' % self.name
        string += 'Variable type:  %s\n' % self.datatype
        string += 'Variable value: %s\n' % self.value
        string += 'Matching chars: %s\n' % self.tomatch
        string += 'Description:    %s\n' % self.description
        return string

    def help_str(self, length):
        """ returns the string [<name> = <value>.... # description] """
        if self.datatype is str:
            valstring = '%s = "%s"' % (self.name, self.value)
        else:
            valstring = '%s = %s' % (self.name, self.value)
        valstring += ' ' + '.' * (length - len(valstring) - 2) + ' '
        return valstring + '# %s' % self.description

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __eq__(self, teststring):
        """ Determines if a variable string matches this variable """

        if len(teststring) > len(self.name):
            return False

        if len(teststring) < self.tomatch:
            return False

        myvar = self.name
        if not self.case_sensitive:
            myvar = self.name.lower()
            teststring = teststring.lower()

        return myvar[:len(teststring)] == teststring

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __ne__(self, teststring):
        """ Not equal """
        return not self.__eq__(teststring)

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def SetValue(self, value):
        """ Sets the value of the variable """
        if self.datatype is str:
            self.value = value.replace('"', '').replace("'", '')
        else:
            self.value = self.datatype(value)


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Namelist(object):
    """ Sets up a namelist. This holds many different Variables, and these
       variables will only be recognized when parsing this particular namelist.
       Set up to mimic the behavior of a Fortran namelist (so that the input is
       similar to the rest of Amber). Some of the known deficiencies:

         o the body of the namelist cannot start on the same line as the start
           or end of the namelist

         o the end of the namelist must be &end or / and must be on its own line

         o It will not (yet) recognize array lengths -- those are simply parsed
           as strings
   """

    MIN_CHARS_TO_MATCH = 4

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __init__(self, trigger, full_name, to_match=3):
        """ Sets up the list of variables, which is just the trigger for now. The
          trigger is a logical variable that gets set to true if this namelist
          is parsed. Open() trips the trigger if it exists. It can be passed in
          as anything that evaluates to False if it doesn't exist.
      """
        self.trigger = trigger
        self.variables = {}
        if self.trigger is not None:
            self.variables = {self.trigger: False}
        self.open = False
        self.full_name = full_name
        self.to_match = to_match

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __eq__(self, nml):
        """ A namelist is equal if the name matches properly """
        return nml == self.full_name[:len(nml)] and len(nml) >= \
               min(self.to_match, len(self.full_name))

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __ne__(self, nml):
        """ Not equal """
        return not self.__eq__(nml)

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def addVariable(self, varname, datatype, default=None, description=None):
        """ Adds a variable to this namelist. It checks to make sure that it's
          going to create a conflict with an existing variable.
        """

        if varname in list(self.variables.keys()):
            raise InternalError('Duplicated variable %s in Namelist' % varname)

        # See if we need to update the minimum number of characters that any
        # variables need based on similarities with this variable
        tomatch = self.MIN_CHARS_TO_MATCH
        for var in list(self.variables.keys()):
            # Trigger never appears, so ignore it
            if var == self.trigger: continue
            hit_difference = False
            for i in range(min(len(var), len(varname))):
                if var[i] == varname[i]: continue
                # At this point, they differ. Update the existing variable if
                # necessary here
                hit_difference = True
                if i + 1 > self.variables[var].tomatch:
                    self.variables[var].tomatch = i + 1
                tomatch = max(tomatch, i + 1)
                break
            # Now handle the case where one variable is a substring of the other
            if not hit_difference:
                if len(varname) > len(var):
                    self.variables[var].tomatch = len(var)
                    tomatch = max(len(var), tomatch)
                else:
                    self.variables[var].tomatch = max(self.variables[var].tomatch,
                                                      len(varname))
                    tomatch = len(varname)
                    break
        self.variables[varname] = Variable(varname, datatype, default, tomatch,
                                           description=description)

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def Open(self):
        """ Signifies that the namelist is open """
        if self.open:
            raise InputError('Namelist already open. Cannot open before closing')

        if self.trigger: self.variables[self.trigger] = True
        self.open = True

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __str__(self):
        """
        Prints out the full contents of this namelist in the Fortran namelist
        format
        """
        maxlen = max([len(v) + len(str(self.variables[v].value))
                      for v in self.variables if v is not self.trigger]) + 7
        maxlen = max(maxlen, 25)
        retstr = '&%s\n' % self.full_name
        for variable in self.variables:
            if variable is self.trigger: continue
            retstr += '  %s\n' % self.variables[variable].help_str(maxlen)
        return retstr + '/'


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class InputFile(object):
    """ Defines the Input File and parses it. You have to add stuff to the parser
       via addNamelist. Use it as follows:

       input = InputFile()

       input.addNamelist('gb', 'gb', [['saltcon', float, 0], ...],
                         trigger='gbrun')
       input.addNamelist('ala', 'alanine_scanning', [['mutant', int, 0]],
                         trigger='alarun')

       INPUT = input.Parse('mmpbsa.in')
   """

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __init__(self):
        """ Initializes the input file, sets up empty arrays/dictionaries """
        self.ordered_namelist_keys = []
        self.namelists = {}
        self.text = ''  # text of the given input file

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def __str__(self):
        """ Prints out the input file """
        if not self.text:
            return

        ret_text = self.text.replace('\n', '\n|')  # Add | to start of each line

        return ('|Input file:\n|---------------------------------------' +
                '-----------------------\n|' + ret_text +
                '-----------------------------------------------------' +
                '---------\n')

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def print_contents(self, destination):
        """ Prints the contents of the input file """
        # Open a file to write to if need be
        if hasattr(destination, 'write'):
            dest = destination
        else:
            dest = open(destination, 'w')
        for namelist in self.ordered_namelist_keys:
            destination.write('%s\n' % self.namelists[namelist])
        # Close the file if we opened it.
        if dest is not destination:
            dest.close()

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def addNamelist(self, name, full_name, variable_list, trigger=None):
        """ Adds a namelist to the input file that will be parsed. Variable list
          must be an array of arrays. Each array element must be an array that
          has the information [varname, datatype, default, chars to match]. The
          'trigger' is the logical variable that gets set to true if this
          namelist is specified in the input file.
        """

        if name in self.ordered_namelist_keys:
            raise InputError('Namelist %s defined multiple times' % name)

        self.ordered_namelist_keys.append(name)
        self.namelists[name] = Namelist(trigger, full_name)

        for var in variable_list:

            if not (isinstance(var, list) or isinstance(var, tuple)) \
                    or len(var) != 4:
                raise InputError('variables in variable_list must be lists of ' +
                                 'length 4. [varname, datatype, default, description]')

            self.namelists[name].addVariable(var[0], var[1], var[2], var[3])

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def _full_namelist_name(self, nml):
        """ Determines what the full namelist name is. We try to make as many
          allowances as possible. We will match the first 3 characters and
          replace all _'s with
      """
        nml = nml.replace(' ', '_')  # replaces spaces with _'s
        for key in self.ordered_namelist_keys:
            if self.namelists[key] == nml: return key

        raise InputError('Unrecognized namelist %s' % nml)

    # -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#

    def Parse(self, filename):
        """
        This subroutine parses the input file. Only data in namelists are
          parsed, and all namelists must be set prior to calling this routine.

          It will create a dictionary of Input variables for all variables in
          all namelists. They all flood the same namespace. If there are any
          conflicts between variables in namelists, an error will be raised.
          Make sure all input variables are unique!
        """
        from os.path import exists

        # Make sure our file exists

        if filename is None:
            raise InputError("No input file was provided!")
        if not exists(filename):
            raise InputError("Can't find input file (%s)" % filename)

        # Load the whole thing into memory. This should be plenty short enough.
        lines = open(filename, 'r').readlines()

        # Save the text of the input file so we can echo it back later
        self.text = ''.join(lines)

        # We will loop through the input file three times:
        #
        # 1st: Load all of the data into an array (namelist_fields)
        # 2nd: Combine multi-element values (to allow commas in input variables)
        # 3rd: Loop through the fields to change the values of the variables.

        declared_namelists = []  # keep track of the namelists we've found so far
        namelist_fields = []  # entries in a given namelist
        innml = False  # are we in a namelist now? Don't enter multiple

        # split up the input file into separate fields by comma

        for line in lines:
            # Skip title lines (we are flexible here) and comments
            if not innml and not line.strip().startswith('&'):
                continue
            if line.strip().startswith('#') or line.strip().startswith('!'):
                continue

            # Catch some errors
            if innml and line.strip().startswith('&'):
                raise InputError('Invalid input. Terminate each namelist prior ' +
                                 'to starting another one.')

            # End of a namelist
            elif innml and line.strip() in ['/', '&end']:
                innml = False

            # Now if we finally find a namelist
            elif not innml and line.strip().startswith('&'):
                innml = True
                namelist = line.strip()[1:].lower()
                namelist = self._full_namelist_name(namelist)

                if namelist in declared_namelists:
                    raise InputError('Namelist %s specified multiple times' %
                                     namelist)

                self.namelists[namelist].Open()
                declared_namelists.append(namelist)
                namelist_fields.append([])

            # We are in a namelist here, now fill in the fields
            elif innml:
                items = line.strip().split(',')

                # Screen any blank fields
                j = 0
                while j < len(items):
                    items[j] = items[j].strip()
                    if len(items[j]) == 0:
                        items.pop(j)
                    else:
                        j += 1

                namelist_fields[len(namelist_fields) - 1].extend(items)

            # end if [elif innml]

        # end for line in lines

        # Combine any multi-element fields into the last field that has a = in it

        begin_field = -1
        for i in range(len(namelist_fields)):
            for j in range(len(namelist_fields[i])):
                if not '=' in namelist_fields[i][j]:
                    if begin_field == -1:
                        raise InputError('Invalid input file! Error reading ' +
                                         'namelist %s' % declared_namelists[i])
                    else:
                        namelist_fields[i][begin_field] += \
                            ',%s' % namelist_fields[i][j]
                else:
                    begin_field = j

        # Now parse through the items to add them to the master dictionary. Note
        # that thanks to the last step, all data in namelist_fields will be
        # contained within fields that have a '='. All others can be ignored

        for i in range(len(namelist_fields)):
            for j in range(len(namelist_fields[i])):

                if not '=' in namelist_fields[i][j]:
                    continue
                else:
                    var = namelist_fields[i][j].split('=')
                    var[0] = var[0].strip()
                    var[1] = var[1].strip()

                    # Now we have to loop through all variables in that namelist to
                    # see if this is the variable we want.

                    found = False
                    for key in list(self.namelists[declared_namelists[i]].variables.keys()):
                        if self.namelists[declared_namelists[i]].variables[key] == \
                                var[0]:
                            self.namelists[declared_namelists[i]].variables[key]. \
                                SetValue(var[1])
                            found = True
                            break

                    if not found:
                        raise InputError('Unknown variable %s in &%s' % (var[0],
                                                                         declared_namelists[i]))

        # Now it's time to fill the INPUT dictionary

        INPUT = {}

        for nml in self.ordered_namelist_keys:
            for var in list(self.namelists[nml].variables.keys()):

                # Here, the triggers are just bool types, so protect from accessing
                # an attribute that doesn't exist! We only allow Variable types and
                # bool types

                var_object = self.namelists[nml].variables[var]
                try:
                    INPUT[var] = self.namelists[nml].variables[var].value
                except AttributeError:
                    if isinstance(var_object, bool):
                        INPUT[var] = var_object
                    else:
                        raise InputError('Disallowed namelist variable type')

        return INPUT


# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

# Define the MM/PBSA input file here
input_file = InputFile()

strip_mask = ':WAT,Cl*,CIO,Cs+,IB,K*,Li+,MG*,Na+,Rb+,CS,RB,NA,F,CL'

# Add namelists with a list of variables. The variables are added to the
# namelists in lists. The entries are:
# [<variable name> <variable type> <default value> <# of characters to match>]

input_file.addNamelist('general', 'general',
                       [
                           ['assign_chainID', int, 0, 'Assign chains ID'],
                           ['debug_printlevel', int, 0, 'Increase debugging info printed'],
                           ['endframe', int, 9999999, 'Last frame to analyze'],
                           ['entropy', int, 0, 'Do quasi-harmonic or interaction entropy calculation'],
                           ['entropy_seg', int, 25, 'Trajectory segment to calculate interaction entropy'],
                           ['entropy_temp', float, 298.15, 'Temperature to calculate interaction entropy'],
                           ['full_traj', int, 0, 'Print a full traj. AND the thread trajectories.'],
                           ['gmx_path', str, '', 'Force to use this path to get GROMACS executable.'],
                           ['interval', int, 1, 'Number of frames between adjacent frames analyzed.'],
                           ['ions_parameters', int, 1, 'Define ions parameters to build the Amber topology.'],
                           ['keep_files', int, 2, 'How many files to keep after successful completion.'],
                           # ['ligand_mask', str, None, 'Amber mask of ligand atoms in complex prmtop.'],
                           ['ligand_forcefield', str, "leaprc.gaff", 'Define the force field to build Amber topology for '
                                                             'ligand (small molecule)'],
                           ['netcdf', int, 0, 'Use NetCDF intermediate trajectories.'],
                           ['overwrite_data', int, 0, 'Defines whether the gmxMMPBSA data will be overwritten'],
                           ['PBRadii', int, 3, 'Define PBRadii to build amber topology from GROMACS files.'],
                           ['protein_forcefield', str, "oldff/leaprc.ff99SB", 'Define the force field to build Amber '
                                                                              'topology for protein'],

                           # ['receptor_mask', str, None, 'Amber mask of receptor atoms in complex prmtop.'],
                           # ['search_path', str, '', 'Look for intermediate programs in all of PATH'],
                           ['solvated_trajectory', int, 1, 'Define if it is necessary to cleanup the trajectories'],
                           ['startframe', int, 1, 'First frame to analyze.'],
                           ['strip_mask', str, strip_mask, 'Amber mask to strip from solvated prmtop'],
                           ['use_sander', int, 0, 'Use sander to compute energies.'],
                           ['verbose', int, 1, 'How many energy terms to print in the final output.']
                       ], trigger=None)

input_file.addNamelist('gb', 'gb',
                       [
                           ['ifqnt', int, 0, 'Use QM on part of the system'],
                           ['igb', int, 5, 'GB model to use'],
                           ['intdiel', float, 1.0, 'Internal dielectric constant for sander'],
                           ['qm_theory', str, '', 'Semi-empirical QM theory to use.'],
                           ['qm_residues', str, '', 'Residues to treat with QM'],
                           ['qmcharge_com', int, 0, 'Charge of QM region in complex'],
                           ['qmcharge_lig', int, 0, 'Charge of QM region in ligand'],
                           ['qmcharge_rec', int, 0, 'Charge of QM region in receptor'],
                           ['qmcut', float, 9999, 'Cutoff in the QM region'],
                           ['saltcon', float, 0, 'Salt concentration (M)'],
                           ['surfoff', float, 0.0, 'Surface tension offset'],
                           ['surften', float, 0.0072, 'Surface tension'],
                           ['molsurf', int, 0, 'Use Connelly surface (\'molsurf\' program)'],
                           ['msoffset', float, 0.0, 'Offset for molsurf calculation.'],
                           ['probe', float, 1.4, 'Solvent probe radius for surface area calc.']
                       ], trigger='gbrun')

input_file.addNamelist('pb', 'pb',
                       [
                           ['cavity_offset', float, -0.5692, 'Offset for nonpolar solvation calc.'],
                           ['cavity_surften', float, 0.0378, 'Surface tension'],
                           ['exdi', float, 80, 'External dielectric constant'],
                           ['emem', float, 1.0, 'Membrane dielectric constant'],
                           ['memopt', int, 0, 'Use PB optimation for membrane'],
                           ['memoptzero', int, 0, 'Used in PB optimization for ligand'],
                           ['sasopt', int, 0, 'Molecular surface in PB implict model'],
                           ['solvopt', int, 2, 'Select iterative solver'],
                           ['mthick', float, 40.0, 'Membrane thickness'],
                           ['mctrdz', float, 0.0, 'Distance to offset membrane in Z direction'],
                           ['maxarcdot', int, 1500, 'Number of dots used to store arc dots per atom '],
                           ['poretype', int, 1, 'Use exclusion region for channel proteins'],
                           ['npbverb', int, 0, 'Option to turn on verbose mode'],
                           ['nfocus', int, 2, 'Electrostatic focusing calculation'],
                           ['bcopt', int, 5, 'Boundary condition option'],
                           ['eneopt', int, 2, 'Compute electrostatic energy and forces'],
                           ['frcopt', int, 0, 'Output for computing electrostatic forces'],
                           ['cutfd', float, 5.0, 'Cutoff for finite-difference interactions'],
                           ['cutnb', float, 0.0, 'Cutoff for nonbonded interations'],
                           ['ipb', int, 2, 'Dielectric model for PB'],
                           ['fillratio', float, 4, 'See "fillratio" in AmberTools/PBSA manual.'],
                           ['indi', float, 1, 'Internal dielectric constant'],
                           ['inp', int, 2, 'Nonpolar solvation method'],
                           ['istrng', float, 0, 'Ionic strength (M)'],
                           ['linit', int, 1000, 'Number of SCF iterations'],
                           ['prbrad', float, 1.4, 'Probe radius'],
                           ['radiopt', int, 1, 'Use optimized radii?'],
                           ['sander_apbs', int, 0, 'Use sander.APBS?'],
                           ['scale', float, 2, 'Distance between grid points (Angstroms)']
                       ], trigger='pbrun')

input_file.addNamelist('ala', 'alanine_scanning',
                       [['mutant_only', int, 0, 'Only compute mutant energies'],
                        ['mutant', str, 'ALA', 'Defines if Alanine or Glycine scanning will be performed'],
                        ['mutant_res', str, None, 'Which residue will be mutated']],
                       trigger='alarun')

input_file.addNamelist('nmode', 'nmode',
                       [
                           ['dielc', float, 1, 'Dielectric constant'],
                           ['drms', float, 0.001, 'Minimization gradient cutoff'],
                           ['maxcyc', int, 10000, 'Maximum number of minimization cycles'],
                           ['nminterval', int, 1, 'Interval to take snapshots for normal mode analysis'],
                           ['nmendframe', int, 1000000, 'Last frame to analyze for normal modes'],
                           ['nmode_igb', int, 1, 'GB model to use for normal mode calculation'],
                           ['nmode_istrng', float, 0, 'Ionic strength for GB model (M)'],
                           ['nmstartframe', int, 1, 'First frame to analyze for normal modes']
                       ], trigger='nmoderun')

input_file.addNamelist('decomp', 'decomposition',
                       [
                           ['csv_format', int, 1, 'Write decomposition data in CSV format'],
                           ['dec_verbose', int, 0, 'Control energy terms are printed to the output'],
                           ['idecomp', int, 0, 'Which type of decomposition analysis to do'],
                           ['print_res', str, 'within 6', 'Which residues to print decomposition data for']
                       ], trigger='decomprun')

input_file.addNamelist('rism', 'rism',
                       [
                           ['buffer', float, 14, 'Distance between solute and edge of grid'],
                           ['closure', str, 'kh', 'Closure equation to use'],
                           ['closureorder', int, 1, 'Order of closure if PSE'],
                           ['grdspc', float, 0.5, 'Grid spacing'],
                           ['ng', str, '-1,-1,-1', 'Number of grid points'],
                           ['polardecomp', int, 0, 'Break solv. energy into polar and nonpolar terms'],
                           ['rism_verbose', int, 0, 'Control how much 3D-RISM info to print'],
                           ['solvbox', str, '-1,-1,-1', 'Box limits'],
                           ['solvcut', float, None, 'Cutoff of the box'],
                           ['thermo', str, 'std', 'Type of thermodynamic analysis to do'],
                           ['tolerance', float, 1.0e-5, 'Convergence tolerance']
                       ], trigger='rismrun')
